import concurrent.futures
import time
import numpy as np
import torch
from cellpose import models
from tqdm import tqdm

from ..utils.segmentation_tracking.mask_utils import (
    centroid,
    from_labeling_with_roi,
    simplify,
)
from ..utils.cell_spot import CellSpot
from ..utils.cell_track import CellTrack
from ..utils.mid_body_detection.spatial_laptrack import (
    SpatialLapTrack,
)
from ..utils.track_generation import generate_tracks_from_spots


class SegmentationTrackingFactory:
    """Class to perform cell segmentation and tracking.

    Parameters
    ----------
    model_path : str
        Path to the cellpose model
    augment : bool
        cf cellpose documentation
    cellprob_threshold : float
        cf cellpose documentation
    flow_threshold : float
        cf cellpose documentation
    gap_closing_max_distance_ratio : float
        Ratio of average spot size to use for gap closing
    linking_max_distance_ratio : float
        Ratio of average spot size
    max_frame_gap : int
        Maximum number of frames to consider for gap closing
    """

    def __init__(
        self,
        model_path: str,
        augment=True,
        cellprob_threshold=0.0,
        flow_threshold=0.0,
        gap_closing_max_distance_ratio=0.5,
        linking_max_distance_ratio=1,
        max_frame_gap=CellTrack.max_frame_gap,
        minimum_cell_track_length=10,
    ) -> None:
        self.model_path = model_path
        self.augment = augment
        self.cellprob_threshold = cellprob_threshold
        self.flow_threshold = flow_threshold
        self.gap_closing_max_distance_ratio = gap_closing_max_distance_ratio
        self.linking_max_distance_ratio = linking_max_distance_ratio
        self.max_frame_gap = max_frame_gap
        self.minimum_cell_track_length = minimum_cell_track_length

    @staticmethod
    def get_spots_from_cellpose(
        cellpose_results: np.ndarray, parallel: bool = False
    ) -> dict[int, list[CellSpot]]:
        """Extract spots from cellpose results.

        Parameters
        ----------
        cellpose_results : np.ndarray
            TYX
        parallel : bool
            If True, use parallel processing.
            Set to False by default, as it is not obvious that this is much faster.

        Returns
        -------
        dict[int, list[CellSpot]]
            Dictionary with frame number as key and list of cell spots as value.
        """

        def get_spots_from_frame(
            frame: int, cellpose_result: np.ndarray
        ) -> tuple[int, list[CellSpot]]:
            """Extract spots from a single frame.

            Returns
            -------
            list[CellSpot]
                List of cell spots.
            """
            # Create and simplify polygons like Trackmate
            # NB: be careful, Trackmate switches x and y
            polygons = from_labeling_with_roi(cellpose_result)
            # assert len(polygons) == cellpose_result.max()
            simplified_polygons = []
            for polygon in polygons:
                simplified_polygon = simplify(polygon, interval=2, epsilon=0.5)
                simplified_polygons.append(simplified_polygon)
            # Get spots from polygons
            cell_spots = []
            for polygon in simplified_polygons:
                # Compute cell bounding box
                abs_min_x, abs_max_x, abs_min_y, abs_max_y = (
                    np.abs(np.min(polygon.y)),
                    np.abs(np.max(polygon.y)),
                    np.abs(np.min(polygon.x)),
                    np.abs(np.max(polygon.x)),
                )
                # Compute cell centroid
                cell_centroid = centroid(
                    polygon.y,
                    polygon.x,
                )  # (x, y)
                cell_spot = CellSpot(
                    frame,
                    cell_centroid[0],  # x
                    cell_centroid[1],  # y
                    -1,
                    abs_min_x,
                    abs_max_x,
                    abs_min_y,
                    abs_max_y,
                    [[x, y] for x, y in zip(polygon.y, polygon.x)],
                )
                cell_spots.append(cell_spot)
            return frame, cell_spots

        if parallel:
            future_list = []
            with concurrent.futures.ThreadPoolExecutor() as e:
                for frame, cellpose_result in enumerate(cellpose_results):
                    future_list.append(
                        e.submit(get_spots_from_frame, frame, cellpose_result)
                    )

            cell_dictionary = {
                res.result()[0]: res.result()[1]
                for res in concurrent.futures.as_completed(future_list)
            }
        else:
            cell_dictionary = {}
            for frame, cellpose_result in enumerate(tqdm(cellpose_results)):
                _, spots = get_spots_from_frame(frame, cellpose_result)
                cell_dictionary[frame] = spots

        # Give id number to cell spots
        id_number = 0
        for frame in range(len(cellpose_results)):
            if frame not in cell_dictionary:
                continue
            for cell in cell_dictionary[frame]:
                cell.id = id_number
                id_number += 1

        return cell_dictionary

    def perform_segmentation(
        self,
        video: np.ndarray,
    ) -> tuple[np.ndarray, list[np.ndarray], float]:
        """Perform cell segmentation using cellpose.

        Parameters
        ----------
        video : np.ndarray
            TCYX

        Returns
        -------
        np.ndarray
            Cellpose results. TYX.
        list[np.ndarray]
            Cellpose flows.
        float
            Expected diameter of the cells.
        """

        # Cellpose segmentation
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        model = models.CellposeModel(
            pretrained_model=self.model_path, device=device
        )

        print("Run cellpose...")
        start = time.time()
        cellpose_results, flows, _ = model.eval(  # TYX
            video,
            channels=[3, 0],
            diameter=0,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            augment=self.augment,
            resample=False,
        )
        time_second = int(time.time() - start)
        print(f"Done in {time_second} seconds.")

        return cellpose_results, flows, model.diam_labels

    def perform_tracking(
        self, cellpose_results: np.ndarray, diam_labels: float
    ) -> tuple[list[CellSpot], list[CellTrack]]:
        """Perform tracking using laptrack.

        Parameters
        ----------
        cellpose_results : np.ndarray
            TYX
        diam_labels : float
            Expected diameter of the cells.

        Returns
        -------
        list[CellSpot]
            List of cell spots.
        list[CellTrack]
            List of cell tracks.
        """
        cell_spots_dictionary = self.get_spots_from_cellpose(cellpose_results)

        tracking_method = SpatialLapTrack(
            spatial_coord_slice=slice(0, 2),
            spatial_metric="euclidean",
            track_dist_metric="euclidean",
            track_cost_cutoff=diam_labels * self.linking_max_distance_ratio,
            gap_closing_dist_metric="euclidean",
            gap_closing_cost_cutoff=diam_labels
            * self.gap_closing_max_distance_ratio,
            gap_closing_max_frame_count=3,
            splitting_cost_cutoff=False,
            merging_cost_cutoff=False,
            alternative_cost_percentile=100,
        )
        cell_tracks = generate_tracks_from_spots(
            cell_spots_dictionary, tracking_method
        )

        # Keep only tracks with a minimum length
        cell_tracks = [
            track
            for track in cell_tracks
            if len(track.spots) >= self.minimum_cell_track_length
        ]

        cell_spots = []
        for frame_spots in cell_spots_dictionary.values():
            cell_spots.extend(frame_spots)

        return cell_spots, cell_tracks

    def perform_segmentation_tracking(
        self,
        video: np.ndarray,
    ) -> tuple[list[CellSpot], list[CellTrack]]:
        """Perform cell segmentation and tracking.

        Parameters
        ----------
        video : np.ndarray
            TCYX

        Returns
        -------
        list[CellSpot]
            List of cell spots.
        list[CellTrack]
            List of cell tracks.
        """

        cellpose_results, _, diam_labels = self.perform_segmentation(video)
        cell_spots, cell_tracks = self.perform_tracking(
            cellpose_results, diam_labels
        )

        return cell_spots, cell_tracks



import torch.nn as nn

from cnn_framework.utils.models.resnet_classifier import ResnetClassifier


class MetaphaseCnnModel(ResnetClassifier):
    """Model for metaphase classification."""

    def __init__(
        self,
        nb_classes: int,
        nb_input_channels: int,
        encoder_name: str,
    ):
        super().__init__(
            nb_classes=nb_classes,
            nb_input_channels=nb_input_channels,
            encoder_name=encoder_name,
        )
        # NB: following is totally unused, but old model defines it
        self.conv = nn.Conv2d(nb_input_channels, 3, 3)



import os
import numpy as np
from tqdm import tqdm

from ..constants.tracking import TIME_RESOLUTION
from ..utils.mt_cut_detection.bridges_mt_model_manager import (
    BridgesMtModelManager,
)
from ..utils.mt_cut_detection.bridges_mt_cnn_model_params import (
    BridgesMtCnnModelParams,
)
from ..utils.tools import apply_hmm, perform_cnn_inference
from ..utils.mt_cut_detection.impossible_detection import (
    ImpossibleDetection,
)
from ..utils.mitosis_track import MitosisTrack


class MtCutDetectionFactory:
    """
    Class to detect first and second MT cut.

    Parameters
    ----------
    margin : int
        Number of pixels on each side of the midbody.
    # Light spot detection parameters
    intensity_threshold_light_spot : int
        Intensity threshold for the light spot detection.
    h_maxima_light_spot : int
        h for the h_maxima function (light spot detection).
    center_tolerance_light_spot : int
        Center tolerance to not count the light spots that are too close to the center.
    min_percentage_light_spot : float
        Minimum percentage of frames with light spots to consider the mitosis as a light spot mitosis.
    crop_size_light_spot : int
        Size of the crop for the light spot detection.
    length_light_spot : int
        Length of the video to check around the mt cut for light spot detection.
    """

    def __init__(
        self,
        margin=50,
        intensity_threshold_light_spot=350,
        h_maxima_light_spot=130,
        center_tolerance_light_spot=5,
        min_percentage_light_spot=0.1,
        crop_size_light_spot=20,
        length_light_spot=3,
    ) -> None:
        self.margin = margin
        self.intensity_threshold_light_spot = intensity_threshold_light_spot
        self.h_maxima_light_spot = h_maxima_light_spot
        self.center_tolerance_light_spot = center_tolerance_light_spot
        self.min_percentage_light_spot = min_percentage_light_spot
        self.crop_size_light_spot = crop_size_light_spot
        self.length_light_spot = length_light_spot

    @staticmethod
    def _is_bridges_classification_impossible(
        mitosis_track: MitosisTrack, video: np.ndarray
    ) -> bool:
        """
        Bridges classification is impossible if:
        - no midbody spots
        - more than 2 daughter tracks
        - nucleus is near border
        - no midbody spot after cytokinesis

        Parameters
        ----------
        mitosis_track : MitosisTrack
            Mitosis track to check.
        video : np.ndarray
            Video of the mitosis track. TYXC.

        Returns
        -------
        bool
            True if classification is impossible, False otherwise.
        """

        if mitosis_track.is_near_border(video):
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.NEAR_BORDER
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.NEAR_BORDER
            )
            return True

        if len(mitosis_track.daughter_track_ids) >= 2:
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.MORE_THAN_TWO_DAUGHTER_TRACKS
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.MORE_THAN_TWO_DAUGHTER_TRACKS
            )
            return True

        if not mitosis_track.mid_body_spots:
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED
            )
            return True

        if (
            max(mitosis_track.mid_body_spots.keys())
            < mitosis_track.key_events_frame["cytokinesis"]
        ):
            mitosis_track.key_events_frame["first_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS
            )
            mitosis_track.key_events_frame["second_mt_cut"] = (
                ImpossibleDetection.NO_MID_BODY_DETECTED_AFTER_CYTOKINESIS
            )
            return True

        return False

    def update_mt_cut_detection(
        self,
        mitosis_tracks: list[MitosisTrack],
        video: np.ndarray,
        hmm_bridges_parameters_file: str,
        bridges_mt_cnn_model_path: str,
        debug_mode=False,
    ) -> dict[str, dict]:
        """
        Update micro-tubules cut detection using bridges classification.

        Parameters
        ----------
        mitosis_tracks : list[MitosisTrack]
            List of mitosis tracks to update.
        video : np.ndarray
            Video of the mitosis tracks. TYXC.
        hmm_bridges_parameters_file : str
            Path to the HMM parameters file.
        bridges_mt_cnn_model_path : str
            Path to the bridges CNN model.
        debug_mode : bool
            If True, return the classified bridges.

        Returns
        -------
        dict[str, dict]
            Classified bridges.
        """
        # Run CNN classification
        classified_bridges = self._classify_bridges(
            mitosis_tracks, video, bridges_mt_cnn_model_path
        )

        print("\nPredictions performed successfully.")

        # Read HMM parameters
        if not os.path.exists(hmm_bridges_parameters_file):
            raise FileNotFoundError(
                f"File {hmm_bridges_parameters_file} not found"
            )
        hmm_parameters = np.load(hmm_bridges_parameters_file)

        # Check if classification is impossible and smooth
        for mitosis_track in tqdm(mitosis_tracks):
            classification_impossible = (
                self._is_bridges_classification_impossible(
                    mitosis_track, video
                )
            )

            if classification_impossible:
                continue

            # Force first frame to be class 0
            classified_bridges["predictions"][mitosis_track.id][0] = 0

            # Correct the sequence with HMM
            classified_bridges["predictions_after_hmm"][mitosis_track.id] = (
                apply_hmm(
                    hmm_parameters,
                    classified_bridges["predictions"][mitosis_track.id],
                )
            )

            # Get index of first element > 0 in sequence (first MT cut)
            first_mt_cut_frame_rel = next(
                (
                    i
                    for i, x in enumerate(
                        classified_bridges["predictions_after_hmm"][
                            mitosis_track.id
                        ]
                    )
                    if x > 0
                ),
                -1,
            )

            # Ignore if no MT cut detected
            if first_mt_cut_frame_rel == -1:
                mitosis_track.key_events_frame["first_mt_cut"] = (
                    ImpossibleDetection.NO_CUT_DETECTED
                )
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.NO_CUT_DETECTED
                )
                continue

            # Ignore if cut is too short, i.e. less than 50 minutes
            first_frame = min(classified_bridges["frames"][mitosis_track.id])
            first_mt_cut_frame_abs = first_frame + first_mt_cut_frame_rel
            if (
                first_mt_cut_frame_abs
                - mitosis_track.key_events_frame["cytokinesis"]
            ) < 50 / TIME_RESOLUTION:
                mitosis_track.key_events_frame["first_mt_cut"] = (
                    ImpossibleDetection.TOO_SHORT_CUT
                )
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.TOO_SHORT_CUT
                )
                continue

            if mitosis_track.light_spot_detected(
                video,
                first_mt_cut_frame_abs,
                self.length_light_spot,
                self.crop_size_light_spot,
                self.h_maxima_light_spot,
                self.intensity_threshold_light_spot,
                self.center_tolerance_light_spot,
                self.min_percentage_light_spot,
            ):
                mitosis_track.key_events_frame["first_mt_cut"] = (
                    ImpossibleDetection.LIGHT_SPOT
                )
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.LIGHT_SPOT
                )
                continue

            # Update mitosis track accordingly
            mitosis_track.key_events_frame["first_mt_cut"] = (
                first_mt_cut_frame_abs
            )

            # Get index of first element > 1 in sequence (second MT cut)
            second_mt_cut_frame_rel = next(
                (
                    i
                    for i, x in enumerate(
                        classified_bridges["predictions_after_hmm"][
                            mitosis_track.id
                        ]
                    )
                    if x > 1
                ),
                -1,
            )

            # get the frame of the second MT cut
            if second_mt_cut_frame_rel == -1:
                mitosis_track.key_events_frame["second_mt_cut"] = (
                    ImpossibleDetection.NO_CUT_DETECTED
                )
                continue

            second_mt_cut_frame_abs = first_frame + second_mt_cut_frame_rel

            # Update mitosis track accordingly
            mitosis_track.key_events_frame["second_mt_cut"] = (
                second_mt_cut_frame_abs
            )

            if debug_mode:
                # In debug, expect only one mitosis track
                return classified_bridges

        return classified_bridges

    def _classify_bridges(
        self,
        mitosis_tracks: list[MitosisTrack],
        video: np.ndarray,
        bridges_mt_cnn_model_path: str,
    ) -> dict[str, dict]:
        """
        Classify bridges using a CNN model.

        Parameters
        ----------
        mitosis_tracks : list[MitosisTrack]
            List of mitosis tracks to classify.
        video : np.ndarray
            Video of the mitosis tracks. TYXC.
        bridges_mt_cnn_model_path : str
            Path to the bridges CNN model.

        Returns
        -------
        dict[str, dict]
            Classified bridges.
        """
        # Get bridge crops
        bridges = {
            "images": {},
            "frames": {},
            "predictions": {},
            "predictions_after_hmm": {},
        }
        for mitosis_track in mitosis_tracks:
            images, frames = mitosis_track.get_bridge_images(
                video, self.margin
            )
            sir_tubulin_images = [
                np.expand_dims(image[0], axis=0) for image in images
            ]
            bridges["images"][mitosis_track.id] = sir_tubulin_images
            bridges["frames"][mitosis_track.id] = frames

        # Perform classification
        flatten_images = [
            image for images in bridges["images"].values() for image in images
        ]
        predictions = perform_cnn_inference(
            model_path=bridges_mt_cnn_model_path,
            images=flatten_images,
            cnn_model_params=BridgesMtCnnModelParams,
            model_manager=BridgesMtModelManager,
        )

        # Fill predictions
        for mitosis_track_id, images in bridges["images"].items():
            bridges["predictions"][mitosis_track_id] = predictions[
                : len(images)
            ]
            predictions = predictions[len(images) :]

        return bridges



import os
from pathlib import Path
import shutil
from typing import Optional

from magicgui import magic_factory
import tempfile

import numpy as np
from skimage import io


from .utils.tools import re_organize_channels

from .widget_functions.segmentation_tracking import perform_tracking
from .widget_functions.mid_body_detection import perform_mid_body_detection
from .widget_functions.mitosis_track_generation import (
    perform_mitosis_track_generation,
)
from .widget_functions.mt_cut_detection import perform_mt_cut_detection
from .widget_functions.save_results import perform_results_saving


def video_whole_process(
    video: np.ndarray,
    video_name: np.ndarray,
    default_model_check_box: bool,
    segmentation_model: str,
    save_check_box: bool,
    movies_save_dir: str,
    spots_dir_name: str,
    tracks_dir_name: str,
    mitoses_dir_name: str,
) -> None:
    """Perform the whole process on a single video."""

    perform_tracking(
        video,
        str(Path(segmentation_model)) if not default_model_check_box else None,
        video_name,
        spots_dir_name,
        tracks_dir_name,
    )
    perform_mitosis_track_generation(
        video,
        video_name,
        spots_dir_name,
        tracks_dir_name,
        mitoses_dir_name,
    )
    perform_mid_body_detection(
        video,
        video_name,
        mitoses_dir_name,
        tracks_dir_name,
        movies_save_dir if save_check_box else None,
        parallel_detection=True,
    )
    perform_mt_cut_detection(video, video_name, mitoses_dir_name)


@magic_factory(
    call_button="Run Whole Process (Single Video)",
    layout="vertical",
    default_model_check_box=dict(
        widget_type="CheckBox",
        text="Use default segmentation model?",
        value=True,
    ),
    segmentation_model=dict(
        widget_type="FileEdit",
        label="If not checked, cellpose segmentation model: ",
    ),
    save_check_box=dict(
        widget_type="CheckBox", text="Save cell divisions movies?", value=False
    ),
    movies_save_dir=dict(
        widget_type="FileEdit",
        label="If checked, directory to save division movies: ",
        mode="d",
    ),
    results_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save results: ",
        mode="d",
    ),
    debug_mode_check_box=dict(
        widget_type="CheckBox",
        text="Debug mode",
        value=False,
    ),
)
def whole_process(
    img_layer: "napari.layers.Image",
    viewer: "napari.Viewer",
    default_model_check_box: bool,
    segmentation_model: str,
    save_check_box: bool,
    movies_save_dir: str,
    results_save_dir: str,
    debug_mode_check_box: bool,
):

    # Create temporary folders
    spots_dir = tempfile.TemporaryDirectory()
    tracks_dir = tempfile.TemporaryDirectory()
    mitoses_dir = tempfile.TemporaryDirectory()

    video = re_organize_channels(img_layer.data)  # TYXC

    video_whole_process(
        video,
        img_layer.name,
        default_model_check_box,
        segmentation_model,
        save_check_box,
        movies_save_dir,
        spots_dir.name,
        tracks_dir.name,
        mitoses_dir.name,
    )

    # Results saving
    perform_results_saving(
        mitoses_dir.name,
        save_dir=results_save_dir,
        video=img_layer.data,
        viewer=viewer,
    )

    if debug_mode_check_box:
        shutil.copytree(
            spots_dir.name, os.path.join(results_save_dir, "spots")
        )
        shutil.copytree(
            tracks_dir.name, os.path.join(results_save_dir, "tracks")
        )
        shutil.copytree(
            mitoses_dir.name, os.path.join(results_save_dir, "mitoses")
        )

    # Delete temporary folders
    spots_dir.cleanup()
    tracks_dir.cleanup()
    mitoses_dir.cleanup()

    print("\nProcess finished with success!")


@magic_factory(
    call_button="Run Whole Process (Folder)",
    layout="vertical",
    raw_data_dir=dict(
        widget_type="FileEdit",
        label="Data folder: ",
        mode="d",
    ),
    default_model_check_box=dict(
        widget_type="CheckBox",
        text="Use default segmentation model?",
        value=True,
    ),
    segmentation_model=dict(
        widget_type="FileEdit",
        label="If not checked, cellpose segmentation model: ",
    ),
    save_check_box=dict(
        widget_type="CheckBox", text="Save cell divisions movies?", value=False
    ),
    movies_save_dir=dict(
        widget_type="FileEdit",
        label="If checked, directory to save division movies: ",
        mode="d",
    ),
    results_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save results: ",
        mode="d",
    ),
)
def whole_process_folder(
    raw_data_dir: str,
    default_model_check_box: bool,
    segmentation_model: str,
    save_check_box: bool,
    movies_save_dir: str,
    results_save_dir: str,
):

    # Create temporary folders
    spots_dir = tempfile.TemporaryDirectory()
    tracks_dir = tempfile.TemporaryDirectory()
    mitoses_dir = tempfile.TemporaryDirectory()

    # Run process on each video
    tiff_files = list(Path(raw_data_dir).rglob("*.tif"))
    for i, tiff_file in enumerate(tiff_files):
        print(
            f"\nProcessing {tiff_file.stem} - Video {i+1}/{len(tiff_files)}\n"
        )
        video = io.imread(tiff_file)  # TYXC
        video_whole_process(
            video,
            tiff_file.stem,
            default_model_check_box,
            segmentation_model,
            save_check_box,
            movies_save_dir,
            spots_dir.name,
            tracks_dir.name,
            mitoses_dir.name,
        )

    # Results saving
    perform_results_saving(mitoses_dir.name, save_dir=results_save_dir)

    # Delete temporary folders
    spots_dir.cleanup()
    tracks_dir.cleanup()
    mitoses_dir.cleanup()

    print("\nProcess finished with success!")


@magic_factory(
    call_button="Run Segmentation and Tracking",
    layout="vertical",
    spots_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save .bin cell spots: ",
        mode="d",
    ),
    tracks_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save .bin cell tracks: ",
        mode="d",
    ),
    default_model_check_box=dict(
        widget_type="CheckBox",
        text="Use default segmentation model?",
        value=True,
    ),
    segmentation_model=dict(
        widget_type="FileEdit",
        label="If not checked, cellpose segmentation model: ",
    ),
)
def segmentation_tracking(
    img_layer: "napari.layers.Image",
    spots_save_dir: str,
    tracks_save_dir: str,
    default_model_check_box: bool,
    segmentation_model: str,
):

    raw_video = re_organize_channels(img_layer.data)  # TYXC

    # Segmentation and tracking
    perform_tracking(
        raw_video,
        str(Path(segmentation_model)) if not default_model_check_box else None,
        img_layer.name,
        spots_save_dir,
        tracks_save_dir,
    )


@magic_factory(
    call_button="Run Mitosis Track Generation",
    layout="vertical",
    spots_load_dir=dict(
        widget_type="FileEdit",
        label="Directory to load .bin cell spots: ",
        mode="d",
    ),
    tracks_load_dir=dict(
        widget_type="FileEdit",
        label="Directory to load .bin cell tracks: ",
        mode="d",
    ),
    mitoses_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save .bin mitoses: ",
        mode="d",
    ),
)
def mitosis_track_generation(
    img_layer: "napari.layers.Image",
    spots_load_dir: str,
    tracks_load_dir: str,
    mitoses_save_dir: Optional[str],
):
    raw_video = re_organize_channels(img_layer.data)  # TYXC

    perform_mitosis_track_generation(
        raw_video,
        img_layer.name,
        spots_load_dir,
        tracks_load_dir,
        mitoses_save_dir,
    )


@magic_factory(
    call_button="Run Mid-body Detection",
    layout="vertical",
    exported_mitoses_dir=dict(
        widget_type="FileEdit",
        label="Directory to load .bin mitoses:",
        mode="d",
    ),
    exported_tracks_dir=dict(
        widget_type="FileEdit",
        label="Directory to load .bin cell tracks: ",
        mode="d",
    ),
    save_check_box=dict(
        widget_type="CheckBox", text="Save cell divisions movies?", value=False
    ),
    movies_save_dir=dict(
        widget_type="FileEdit",
        label="If checked, directory to save division movies: ",
        mode="d",
    ),
)
def mid_body_detection(
    img_layer: "napari.layers.Image",
    exported_mitoses_dir: str,
    exported_tracks_dir: str,
    save_check_box: bool,
    movies_save_dir: str,
):
    raw_video = re_organize_channels(img_layer.data)  # TYXC
    perform_mid_body_detection(
        raw_video,
        img_layer.name,
        exported_mitoses_dir,
        exported_tracks_dir,
        movies_save_dir if save_check_box else None,
        parallel_detection=True,
    )


@magic_factory(
    call_button="Run MT Cut Detection",
    layout="vertical",
    exported_mitoses_dir=dict(
        widget_type="FileEdit",
        label="Directory to load .bin mitoses: ",
        mode="d",
    ),
)
def micro_tubules_cut_detection(
    img_layer: "napari.layers.Image", exported_mitoses_dir: str
):
    raw_video = re_organize_channels(img_layer.data)  # TYXC
    perform_mt_cut_detection(
        raw_video,
        img_layer.name,
        exported_mitoses_dir,
    )


@magic_factory(
    call_button="Save results",
    layout="vertical",
    exported_mitoses_dir=dict(
        widget_type="FileEdit",
        label="Directory to load .bin mitoses: ",
        mode="d",
    ),
    results_save_dir=dict(
        widget_type="FileEdit",
        label="Directory to save results: ",
        mode="d",
    ),
)
def results_saving(
    img_layer: "napari.layers.Image",
    viewer: "napari.Viewer",
    exported_mitoses_dir: str,
    results_save_dir: str,
):
    perform_results_saving(
        exported_mitoses_dir,
        save_dir=results_save_dir,
        video=img_layer.data,
        viewer=viewer,
    )



from .box_dimensions import BoxDimensions


class BoxDimensionsDln(BoxDimensions):
    """A class to store the dimensions of a box and the delaunay triangulation of the box."""

    def __init__(self):
        super().__init__()
        self.dln = None



""" A module that relies on the Track Abstract Base Class in order to
generate specialized tracks from points.

See Track documentation to learn how to create new specialized Tracks/Spots
"""

from typing import Tuple, TypeVar

import pandas as pd
from laptrack import LapTrack

from .track import Track
from .spot import Spot
from .mid_body_track import MidBodyTrack
from .cell_track import CellTrack
from .mid_body_spot import MidBodySpot
from .cell_spot import CellSpot

SPOT_AND_TRACK_MAPPING = {
    MidBodySpot: MidBodyTrack,
    CellSpot: CellTrack,
}

T = TypeVar("T", *SPOT_AND_TRACK_MAPPING.values())
S = TypeVar("S", *SPOT_AND_TRACK_MAPPING.keys())


def generate_tracks_from_spots(
    spot_dict: dict[int, list[S]], method: LapTrack
) -> list[T]:
    """Generate a list of specialized tracks based on the underlying kind
    of spot.
    Spots must all be of the same type according to isinstance (more specifically,
    they must all share the same type as the first one).

    Parameters
    ----------
    spot_dict : dict[int, list[S]]
        A dictionary mapping frames to a list of spots.
    method : LapTrack
        The tracking method to use.

    Returns
    -------
    list[T]
        A list of specialized tracks.
    """

    if is_spot_dict_empty(spot_dict):
        return []

    inferred_spot_kind: S = validate_inferred_spot_kind(
        infer_spot_kind(spot_dict)
    )
    inferred_track_kind: T = infer_specialized_track_kind(inferred_spot_kind)

    spot_df = convert_spots_to_spotdf(spot_dict, inferred_spot_kind)
    track_df, _, _ = apply_tracking(inferred_spot_kind, spot_df, method)
    track_list = inferred_track_kind.track_df_to_track_list(
        track_df, spot_dict
    )

    return track_list


def is_spot_dict_empty(spot_dict: dict[int, list[Spot]]) -> bool:
    """Checks if a spot_dict is empty, meaning that all values are empty lists

    Parameters
    ----------
    spot_dict : dict[int, list[Spot]]
        A dictionary mapping frames to a list of spots

    Returns
    -------
    bool
        True if the spot_dict is empty, False otherwise
    """
    if len(spot_dict.values()) == 0:
        return True

    for v in spot_dict.values():
        if len(v) > 0:
            return False

    return True


def validate_inferred_spot_kind(inference: type) -> S:
    """Validates the inference of the kind of spot.

    Parameters
    ----------
    inference : type
        The inferred kind of spot

    Returns
    -------
    S
        The kind of spot if it is valid
    """
    if inference in SPOT_AND_TRACK_MAPPING:
        return inference
    raise RuntimeError(
        f"Tracks can only be built from {SPOT_AND_TRACK_MAPPING.keys()}, encountered {inference} instead"
    )


def infer_spot_kind(spot_dict: dict[int, list[Spot]]) -> type:
    """Infers the kind of spot from the first non-empty list of spots in the
    spot_dict.

    Parameters
    ----------
    spot_dict : dict[int, list[Spot]]
        A dictionary mapping frames to a list of spots

    Returns
    -------
    type
        The kind of spot
    """
    for v in spot_dict.values():
        if len(v) != 0:
            t = type(v[0])
            if t is None:
                raise RuntimeError(
                    "None in v[0] found in spot_dict:\n{spot_dict}"
                )
            return t

    # If we land here, it means that spot_dict.values is empty or only contains empty point lists.
    # This case must be handled earlier
    raise RuntimeError("No values in dict")


def infer_specialized_track_kind(spot_kind: type[Spot]) -> type[Track]:
    """Infers the specialized type of track from the kind of Spot.
    If no specialized kind of track is associated with the kind of
    spot, this function raises an error.

    Parameters
    ----------
    spot_kind : type[Spot]
        The kind of spot

    Returns
    -------
    type[Track]
        The specialized kind of track
    """
    kind_mapping = {MidBodySpot: MidBodyTrack, CellSpot: CellTrack}
    inferred_specialized_track_kind = kind_mapping.get(spot_kind)
    if infer_specialized_track_kind is None:
        raise RuntimeError(
            f"No known specialized track is associated with {type(spot_kind)}"
        )
    return inferred_specialized_track_kind


def convert_spots_to_spotdf(
    spot_dict: dict[int, list[Spot]], spot_kind: type[Spot]
) -> pd.DataFrame:
    """Converts a dictionary of spots to a DataFrame.

    Parameters
    ----------
    spot_dict : dict[int, list[Spot]]
        A dictionary mapping frames to a list of spots
    spot_kind : type[Spot]
        The kind of spot

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the spots
    """

    cols = [
        "frame",
        "x",
        "y",
        "idx_in_frame",
    ]

    cols.extend(spot_kind.get_extra_features_name())

    spot_df = pd.DataFrame({c: [] for c in cols})

    for frame, spots in spot_dict.items():
        if len(spots) == 0:
            spot_df.loc[len(spot_df.index)] = [
                frame,
                *[
                    None for _ in range(len(cols) - 1)
                ],  # fills the rest of the cols with None
            ]
        else:
            for idx, spot in enumerate(spots):
                assert (
                    spot.frame == frame
                ), "spot.frame and frame cols must be the same"
                features = [spot.frame, spot.x, spot.y, idx]
                features.extend(spot.get_extra_coordinates())
                spot_df.loc[len(spot_df.index)] = features

    # Cast frame to int for laptrack
    spot_df["frame"] = spot_df["frame"].astype(int)

    return spot_df


def apply_tracking(
    spot_kind: type[Spot],
    spot_df: pd.DataFrame,
    method: LapTrack,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Applies the tracking method to the spot DataFrame.

    Parameters
    ----------
    spot_kind : type[Spot]
        The kind of spot
    spot_df : pd.DataFrame
        A DataFrame containing the spots
    method : LapTrack
        The tracking method to use

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]
        A tuple containing the DataFrame of the tracks, the DataFrame of the
        tracks' features, and the DataFrame of the tracks' links
    """
    coord_cols = ["x", "y"]
    coord_cols.extend(spot_kind.get_extra_features_name())

    tuple_df = method.predict_dataframe(
        spot_df,
        coord_cols,
        only_coordinate_cols=False,
    )

    return tuple_df



"""Playground to run cell tracking."""

import os
import pickle
from typing import Optional
import matplotlib.pyplot as plt
import numpy as np

from cut_detector.data.tools import get_data_path
from cut_detector.factories.segmentation_tracking_factory import (
    SegmentationTrackingFactory,
)


def main(
    segmentation_results_path: Optional[str] = os.path.join(
        get_data_path("segmentation_results"), "example_video.bin"
    ),
):
    """
    Parameters
    ----------
    segmentation_results_path : str
        Path to the segmentation results, to avoid executing cellpose here.
    """
    # Load Cellpose results
    with open(segmentation_results_path, "rb") as f:
        cellpose_results = pickle.load(f)

    # Perform tracking from Cellpose results
    factory = SegmentationTrackingFactory("")
    cell_spots, cell_tracks = factory.perform_tracking(cellpose_results, 171)

    # Display results
    for frame, cellpose_result in enumerate(cellpose_results):
        frame_cells = [cell for cell in cell_spots if cell.frame == frame]
        plt.imshow(cellpose_result)
        for local_cell in frame_cells:
            points = np.array(local_cell.spot_points)
            plt.plot(
                points[:, 0],
                points[:, 1],
                "o",
            )
            plt.plot(
                local_cell.x,
                local_cell.y,
                "x",
            )
        plt.show()


if __name__ == "__main__":
    main()



from __future__ import annotations

from copy import deepcopy
from typing import Optional, Tuple, Union
import numpy as np
from skimage.morphology import extrema, opening
import xmltodict
from munch import Munch
from scipy import ndimage

from ..constants.annotations import (
    NAMES_DICTIONARY,
    get_class_ids_after_first_mt_cut,
    get_class_ids_after_second_mt_cut,
    get_class_ids_after_first_membrane_cut,
)
from ..constants.tracking import (
    CYTOKINESIS_DURATION,
    FRAMES_AROUND_METAPHASE,
    METAPHASE_INDEX,
)
from .mid_body_spot import MidBodySpot
from .cell_track import CellTrack
from .box_dimensions_dln import BoxDimensionsDln
from .box_dimensions import BoxDimensions
from .mt_cut_detection.impossible_detection import ImpossibleDetection
from .image_tools import resize_image, smart_cropping


def cell_counter_frame_to_video_frame(
    cell_counter_frame: int, nb_channels=4
) -> int:
    """
    Cell counter index starts at 1, just like Fiji.

    To count frame, it just concatenates all channels.
    For example, with 4 channels, frames 1, 2, 3 and 4 will be frame 1,
    frames 5, 6, 7 and 8 will be frame 2, etc.
    """
    return (cell_counter_frame - 1) // nb_channels


def snake_to_normal(snake_str: str) -> str:
    """Convert a snake string to a normal string.

    Parameters
    ----------
    snake_str : str

    Returns
    -------
    normal_case_str : str
    """

    # Split the string by underscores
    words = snake_str.split("_")

    # Capitalize the first letter of the first word and make the rest lowercase
    normal_case_str = " ".join(words).capitalize()

    return normal_case_str


class MitosisTrack:
    """
    A class to store the information of a mitosis track.

    Parameters
    ----------
    mother_track_id : int
        Mother track id
    daughter_track_id : int
        Daughter track id
    metaphase_frame : int
        Metaphase frame
    """

    def __init__(
        self,
        mother_track_id: int,
        daughter_track_id: int,
        metaphase_frame: int,
    ):
        # Elementary information
        self.mother_track_id = mother_track_id
        self.daughter_track_ids = [daughter_track_id]
        self.id: Optional[int] = None

        self.metaphase_frame = metaphase_frame

        # Key events: metaphase/cytokinesis/first_mt_cut/second_mt_cut/first_membrane_cut
        # Absolute frame
        self.key_events_frame: dict[str, Union[int, ImpossibleDetection]] = {}
        self.gt_key_events_frame: Optional[dict[str, int]] = None

        # Time
        self.min_frame: Optional[int] = None
        self.max_frame: Optional[int] = None

        # Position
        self.position = BoxDimensions()

        # Delaunay triangulation, by frame
        self.dln_positions: dict[int, BoxDimensionsDln] = {}

        # Used for matching between ground truth and prediction
        self.matched = False

        # Mid body spot indexed by absolute frame
        self.mid_body_spots: dict[int, MidBodySpot] = {}
        self.gt_mid_body_spots: Optional[dict[int, MidBodySpot]] = None

    def is_same_mitosis(
        self, mother_track_id: int, metaphase_frame: int
    ) -> bool:
        """Check if current mitosis is the same as the one given in parameters.

        Parameters
        ----------
        mother_track_id : int
            Mother track id
        metaphase_frame : int
            Metaphase frame

        Returns
        -------
        bool
            True if current mitosis is the same as the one given in parameters
        """
        return (
            self.mother_track_id == mother_track_id
            and self.metaphase_frame == metaphase_frame
        )

    def add_daughter_track(self, daughter_track_id: int) -> None:
        """Add daughter track id to current mitosis track.

        Parameters
        ----------
        daughter_track_id : int
            Daughter track id
        """
        self.daughter_track_ids.append(daughter_track_id)

    def get_mother_daughters_tracks(
        self, tracks: list[CellTrack]
    ) -> Tuple[CellTrack, list[CellTrack]]:
        """Get mother and daughter tracks of current mitosis.

        Parameters
        ----------
        tracks : list[CellTrack]
            List of all tracks in the video

        Returns
        -------
        CellTrack
            Mother track
        list[CellTrack]
            Daughter tracks
        """
        mother_track = [
            track for track in tracks if track.track_id == self.mother_track_id
        ][0]
        daughter_tracks = [
            track
            for track in tracks
            if track.track_id in self.daughter_track_ids
        ]
        return mother_track, daughter_tracks

    def _add_dln_position(
        self, frame: int, frame_dimensions: BoxDimensionsDln
    ) -> None:
        """Add Delaunay position.

        Parameters
        ----------
        frame : int
            Frame number
        frame_dimensions : BoxDimensionsDln
            Delaunay box dimensions
        """
        self.dln_positions[frame] = deepcopy(frame_dimensions)
        # Update absolute min and max accordingly
        self.position.update_from_box_dimensions(frame_dimensions)

    def update_mitosis_start_end(
        self,
        cell_tracks: list[CellTrack],
        mitosis_tracks: list[MitosisTrack],
    ) -> None:
        """Update min and max frame of current mitosis.

        Parameters
        ----------
        cell_tracks : list[CellTrack]
            List of all tracks in the video
        mitosis_tracks : list[MitosisTrack]
            List of all mitosis tracks in the video
        """
        # Get all tracks involved in current mitosis
        mother_track, daughter_tracks = self.get_mother_daughters_tracks(
            cell_tracks
        )

        # Get min and max frame of current mitosis
        # Min is the metaphase frame minus FRAMES_AROUND_METAPHASE, protected against frames before start of mother track
        min_frame = max(
            mother_track.start,
            self.metaphase_frame - FRAMES_AROUND_METAPHASE,
        )
        # For each daughter track, the end is the end of the track OR the next metaphase event of this track
        max_frame = mother_track.stop
        for track in [mother_track] + daughter_tracks:
            track_end_frame = track.stop
            for track_to_merge_bis in mitosis_tracks:
                if (
                    track_to_merge_bis.mother_track_id == track.track_id
                    and track_to_merge_bis.metaphase_frame
                    > self.metaphase_frame  # other mitosis should be strictly after
                ):
                    track_end_frame = min(
                        track_end_frame, track_to_merge_bis.metaphase_frame
                    )
            max_frame = min(max_frame, track_end_frame)

        # Update mitosis_track
        self.min_frame = min_frame
        self.max_frame = max_frame

    def is_near_border(
        self, raw_video: np.ndarray, minimum_distance=20
    ) -> bool:
        """Check if the mitosis is near the border of the video.

        Parameters
        ----------
        raw_video: np.ndarray
            TYXC

        Returns
        -------
        bool
            True if the mitosis is near the border.
        """

        max_height, max_width = raw_video.shape[1], raw_video.shape[2]

        cyto_frame = self.key_events_frame["cytokinesis"]
        last_frame = cyto_frame + CYTOKINESIS_DURATION

        # Get mitosis coordinates between cyto_frame and last_frame
        min_dist_to_border = np.inf
        for frame in range(cyto_frame, last_frame + 1):
            if frame not in self.mid_body_spots:
                continue

            # Get mid-body coordinates
            mid_body_frame = self.mid_body_spots[frame]
            x_rel = mid_body_frame.x
            y_rel = mid_body_frame.y

            x_abs = x_rel + self.position.min_x
            y_abs = y_rel + self.position.min_y
            mid_body_coordinates = (x_abs, y_abs)

            # Get distance to border
            min_x = min(
                mid_body_coordinates[0], max_width - mid_body_coordinates[0]
            )
            min_y = min(
                mid_body_coordinates[1], max_height - mid_body_coordinates[1]
            )

            min_dist_to_border = min(min_dist_to_border, min_x, min_y)

        return min_dist_to_border < minimum_distance

    def update_key_events_frame(self, cell_tracks: list[CellTrack]) -> None:
        """Update key events frame for current mitosis.

        Parameters
        ----------
        cell_tracks : list[CellTrack]
            List of all tracks in the video
        """
        # Get all tracks involved in current mitosis
        mother_track, daughter_tracks = self.get_mother_daughters_tracks(
            cell_tracks
        )

        # Store first metaphase frame
        for frame in range(self.metaphase_frame, mother_track.start, -1):
            # Some frames may be missing since gap closing is allowed
            if frame not in mother_track.spots:
                continue
            if mother_track.spots[frame].predicted_phase != METAPHASE_INDEX:
                self.key_events_frame["metaphase"] = frame + 1
                break

        # If no metaphase frame found, consider it is the first frame of the mother track
        if "metaphase" not in self.key_events_frame:
            self.key_events_frame["metaphase"] = mother_track.start

        # Store first cytokinesis frame - considered as the first frame of daughter tracks
        self.key_events_frame["cytokinesis"] = min(
            [track.start for track in daughter_tracks]
        )

    def update_mitosis_position_dln(
        self, cell_tracks: list[CellTrack]
    ) -> None:
        """
        Update positions of mitosis for each frame and Delaunay triangulation.

        Parameters
        ----------
        cell_tracks : list[CellTrack]
            List of all tracks in the video
        """

        min_frame, max_frame = self.min_frame, self.max_frame
        mother_track, daughter_tracks = self.get_mother_daughters_tracks(
            cell_tracks
        )

        previous_box_dimensions_dln = None
        for frame in range(min_frame, max_frame + 1):
            box_dimensions_dln = mother_track.compute_dln_from_tracks(
                frame,
                previous_box_dimensions_dln,
                additional_tracks=daughter_tracks,
            )
            # Store in case next frame is missing
            previous_box_dimensions_dln = box_dimensions_dln
            # Update accordingly
            self._add_dln_position(frame, box_dimensions_dln)

    def generate_video_movie(
        self, raw_video: np.ndarray
    ) -> Tuple[np.array, np.array]:
        """
        Generate mitosis movie and mask movie from raw video.

        Parameters
        ----------
        raw_video : np.ndarray
            initial video, TYXC

        Returns
        ----------
        mitosis_movie : np.ndarray
            mitosis movie, TYXC
        mask_movie : np.ndarray
            mask movie, TYX
        """

        mitosis_movie, mask_movie = [], []
        for frame in range(self.min_frame, self.max_frame + 1):
            # Get useful data for current frame
            min_x = self.dln_positions[frame].min_x
            max_x = self.dln_positions[frame].max_x
            min_y = self.dln_positions[frame].min_y
            max_y = self.dln_positions[frame].max_y
            dln = self.dln_positions[frame].dln

            # Extract frame image, big enough to keep all spots for current track
            frame_image = raw_video[
                frame,
                self.position.min_y : self.position.max_y,
                self.position.min_x : self.position.max_x,
                :,
            ]  # YXC

            # Generate mask with Delaunay triangulation
            current_frame_shape = (
                max_y - min_y,
                max_x - min_x,
            )  # current spot
            indices = np.stack(np.indices(current_frame_shape), axis=-1)
            out_idx = np.nonzero(dln.find_simplex(indices) + 1)
            single_channel_mask = np.zeros(current_frame_shape)
            single_channel_mask[out_idx] = 1

            # Construct mask image
            mask_image = np.stack(
                [single_channel_mask] * raw_video.shape[-1], axis=0
            )  # CYX
            mask_image = resize_image(
                mask_image,
                method="zero",
                pad_margin_h=[
                    min_y - self.position.min_y,
                    self.position.max_y - max_y,
                ],
                pad_margin_w=[
                    min_x - self.position.min_x,
                    self.position.max_x - max_x,
                ],
            )[
                0, ...
            ]  # YX

            mitosis_movie.append(frame_image)
            mask_movie.append(mask_image)

        mitosis_movie = np.array(mitosis_movie)  # TYXC
        mask_movie = np.array(mask_movie)  # TYX

        return mitosis_movie, mask_movie

    def is_possible_match(self, other_track: MitosisTrack) -> bool:
        """
        Check if two tracks are a possible match. Other track is typically a ground truth track.
        Match is possible if there is an overlap between the two tracks,
        and other track starts no earlier/no later than FRAMES_AROUND_METAPHASE around self start.

        Parameters
        ----------
        other_track : MitosisTrack
            Other track to compare with.

        Returns
        -------
        bool
            True if match is possible.
        """
        if (
            abs(other_track.metaphase_frame - self.metaphase_frame)
            > FRAMES_AROUND_METAPHASE
        ):
            return False

        return self.position.overlaps(other_track.position)

    def add_mid_body_movie(
        self, mitosis_movie: np.ndarray, mask_movie: np.ndarray
    ) -> np.ndarray:
        """
        Add mid-body to mitosis movie.

        Parameters
        ----------
        mitosis_movie : np.ndarray
            TYXC
        mask_movie : np.ndarray
            TYX

        Returns
        ----------
        spots_video : np.ndarray
            TYX C=1
        """

        video_shape = mitosis_movie.shape[:3]
        spots_video = np.zeros(video_shape)  # TYX

        for absolute_frame, spot in self.mid_body_spots.items():
            # Create 1 circle around spot position
            square_size = 2
            spots_video[
                absolute_frame - self.min_frame,
                spot.y - square_size : spot.y + square_size,
                spot.x - square_size : spot.x + square_size,
            ] = 1

        # Add empty dimension at end
        mid_body_movie = np.expand_dims(spots_video, axis=-1)  # TYX C=1

        # Mix mid-body and mask movie
        mask_movie = np.expand_dims(mask_movie, axis=-1)  # TYX C=1
        mid_body_mask_movie = mask_movie + mid_body_movie  # TYX C=1

        # Cast mid_body_mask_movie to mitosis_movie dtype
        mid_body_mask_movie = mid_body_mask_movie.astype(mitosis_movie.dtype)

        mitosis_movie = np.concatenate(
            [mitosis_movie, mid_body_mask_movie], axis=-1
        )  # TYX C=C+1

        return mitosis_movie

    def update_mid_body_ground_truth(
        self, annotation_file: str, nb_channels: int
    ) -> None:
        """
        Update mid body ground truth from CellCounter annotations.

        Parameters
        ----------
        annotation_file : str
            .xml file with annotations from CellCounter
        nb_channels : int
            Number of channels in mitosis movie (very likely to be 4)
        """

        # Initialize gt_key_events_frame - first two events are shared
        self.gt_key_events_frame = {
            "metaphase": self.key_events_frame["metaphase"],
            "cytokinesis": self.key_events_frame["cytokinesis"],
        }
        self.gt_mid_body_spots = {}

        # Read data
        with open(annotation_file) as fd:
            doc = Munch.fromDict(xmltodict.parse(fd.read()))

        for i, type_data in enumerate(
            doc.CellCounter_Marker_File.Marker_Data.Marker_Type
        ):
            assert i == int(type_data.Type) - 1  # order must be kept
            # Ignore if no data
            if "Marker" not in type_data:
                continue
            markers = type_data.Marker
            if not isinstance(markers, list):
                markers = [markers]
            # Sort markers by frame
            markers = sorted(markers, key=lambda x: int(x.MarkerZ))
            for marker in markers:
                x_pos, y_pos, frame = (
                    int(marker.MarkerX),
                    int(marker.MarkerY),
                    cell_counter_frame_to_video_frame(
                        int(marker.MarkerZ), nb_channels
                    ),
                )
                # Create associated spot
                self.gt_mid_body_spots[frame + self.min_frame] = MidBodySpot(
                    frame, x=x_pos, y=y_pos
                )

            # If Name is missing of wrong, assume it is i
            if (
                "Name" not in type_data
                or type_data.Name not in NAMES_DICTIONARY
            ):
                class_index = i
            else:
                class_index = NAMES_DICTIONARY[type_data.Name]
                assert class_index == i

            class_first_frame = cell_counter_frame_to_video_frame(
                int(markers[0].MarkerZ), nb_channels
            )
            class_abs_first_frame = class_first_frame + self.min_frame

            # First MT cut
            if (
                "first_mt_cut" not in self.gt_key_events_frame
                and class_index in get_class_ids_after_first_mt_cut()
            ):
                assert (
                    class_abs_first_frame
                    >= self.gt_key_events_frame["cytokinesis"]
                )  # after metaphase
                self.gt_key_events_frame["first_mt_cut"] = (
                    class_abs_first_frame
                )

            # Second MT cut
            if (
                "second_mt_cut" not in self.gt_key_events_frame
                and class_index in get_class_ids_after_second_mt_cut()
            ):
                assert (
                    class_abs_first_frame
                    >= self.gt_key_events_frame["first_mt_cut"]
                )  # after first MT cut
                self.gt_key_events_frame["second_mt_cut"] = (
                    class_abs_first_frame
                )

            # First membrane cut
            if (
                "first_membrane_cut" not in self.gt_key_events_frame
                and class_index in get_class_ids_after_first_membrane_cut()
            ):
                assert (
                    class_abs_first_frame
                    >= self.gt_key_events_frame["first_mt_cut"]
                )  # after first MT cut
                self.gt_key_events_frame["first_membrane_cut"] = (
                    class_abs_first_frame
                )

    def evaluate_mid_body_detection(
        self, tolerance=10, percent_seen=0.9, avg_as_int: bool = True
    ) -> Tuple[bool, float, Union[int, float]]:
        """
        Mid_body is considered as detected if during at least percent_seen % of frames
        between cytokinesis and second MT cut it is at most tolerance pixels away
        from ground truth during this interval.

        Parameters
        ----------
        tolerance : int
            Maximum distance between ground truth and prediction to consider a match.
        percent_seen : float
            Minimum percentage of frames where mid_body is seen to consider a match.
        avg_as_int : bool
            If True, average position difference is returned as int.

        Returns
        -------
        bool
            True if mid_body is correctly detected
        float
            Percentage of frames where mid_body is detected
        Union[int, float]
            Average position difference between ground truth and prediction
        """

        position_difference = []

        # Check frames until second MT cut or end of annotations
        max_frame = (
            self.gt_key_events_frame["second_mt_cut"]
            if "second_mt_cut" in self.gt_key_events_frame
            else max(self.gt_mid_body_spots.keys())
        )

        for frame in range(self.gt_key_events_frame["cytokinesis"], max_frame):
            if frame not in self.gt_mid_body_spots:
                continue
            if frame not in self.mid_body_spots:
                position_difference.append(1e3)  # random huge value
                continue
            position_difference.append(
                self.gt_mid_body_spots[frame].distance_to(
                    self.mid_body_spots[frame]
                )
            )

        assert len(position_difference) != 0, "No GT points found"

        # Get percent_seen th percentile of position difference
        position_difference = np.array(position_difference)
        max_position_difference = np.quantile(
            position_difference, percent_seen
        )

        is_correctly_detected = max_position_difference < tolerance
        position_difference_wo_outliers = np.array(
            [pos for pos in position_difference if pos < 1e3]
        )
        percent_detected = int(
            (
                len(position_difference_wo_outliers)
                / len(position_difference)
                * 100
            )
        )

        if avg_as_int:
            average_position_difference = int(
                (
                    position_difference_wo_outliers.mean()
                    if len(position_difference_wo_outliers) > 0
                    else 1e3
                )
            )
        else:
            average_position_difference = (
                position_difference_wo_outliers.mean()
                if len(position_difference_wo_outliers) > 0
                else 1e3
            )

        return (
            is_correctly_detected,
            percent_detected,
            average_position_difference,
        )

    def light_spot_detected(
        self,
        video: np.ndarray,
        first_cut_frame: int,
        length_light_spot: int,
        crop_size_light_spot: int,
        h_maxima_light_spot: int,
        intensity_threshold_light_spot: int,
        center_tolerance_light_spot: int,
        min_percentage_light_spot: float,
        print_enabled=False,
    ) -> bool:
        """
        Check if there is a light spot in crops of size crop_size_light_spot
        around the mid-body, in length_light_spot frames around the first micro-tubules cut.

        Spots are detected using h-maxima method with h=h_maxima_light_spot.
        Ignore spots with intensity lower than intensity_threshold_light_spot.
        Ignore spots close to the center (potential mid-bodies), i.e. within
        center_tolerance_light_spot pixels.

        Light spot is considered as detected if at least in min_percentage_light_spot % of frames.

        Parameters
        ----------
        video : np.ndarray
            TYXC
        first_cut_frame : int
            frame of first micro-tubules cut
        length_light_spot : int
            number of frames around first_cut_frame to consider
        crop_size_light_spot : int
            size of crop around mid-body to consider
        h_maxima_light_spot : int
            h parameter of h-maxima method
        intensity_threshold_light_spot : int
            minimum intensity of spot to consider
        center_tolerance_light_spot : int
            tolerance around mid-body to ignore spots
        min_percentage_light_spot : float
            minimum percentage of frames where light spot is detected
        print_enabled : bool
            if True, print intermediate results

        Returns
        -------
        spot_detected : bool
            True if light spot is detected
        """
        # Get the mitosis video crop
        video = video[
            :,
            self.position.min_y : self.position.max_y,
            self.position.min_x : self.position.max_x,
            :,
        ]

        nb_spot_detected = 0
        frame_counted = 0
        # Iterate over video frames
        for i in range(-length_light_spot // 2, length_light_spot // 2):
            frame = first_cut_frame + i

            # Make sure mid-body exists at frame
            if not self.mid_body_spots or frame not in self.mid_body_spots:
                continue

            # Get mid-body coordinates
            mid_body_frame = self.mid_body_spots[frame]
            x_pos, y_pos = mid_body_frame.x, mid_body_frame.y

            # Extract image and crop on the midbody
            img = np.transpose(video[frame, ...], (2, 0, 1))  # CYX
            crop = smart_cropping(
                img, crop_size_light_spot, x_pos, y_pos, pad=True
            )[
                0, ...
            ]  # YX

            # Perform opening to remove small spots and apply h_maxima to get potential spots
            filtered_image = opening(crop, footprint=np.ones((3, 3)))
            local_maxima = extrema.h_maxima(
                filtered_image, h_maxima_light_spot
            )

            # Label spot regions and remove inconsistent ones
            labeled_local_maxima, nb_labels = ndimage.label(
                local_maxima, structure=np.ones((3, 3))
            )
            for label in range(1, nb_labels + 1):
                # Labels intensity in original image has to be higher than threshold
                if (
                    crop[np.where(labeled_local_maxima == label)].mean()
                    < intensity_threshold_light_spot
                ):
                    labeled_local_maxima[labeled_local_maxima == label] = 0
            # Re-label accordingly
            labeled_local_maxima, nb_labels = ndimage.label(
                labeled_local_maxima > 0, structure=np.ones((3, 3))
            )

            # Get center of mass to locate spots
            spots = ndimage.center_of_mass(
                local_maxima, labeled_local_maxima, range(1, nb_labels + 1)
            )
            spots = np.asarray(spots, dtype=np.int64)

            # Remove spots that are too close to the center
            for spot in spots:
                if (
                    np.abs(spot[0] - crop_size_light_spot)
                    < center_tolerance_light_spot
                ) and (
                    np.abs(spot[1] - crop_size_light_spot)
                    < center_tolerance_light_spot
                ):
                    spots = np.delete(
                        spots, np.where((spots == spot).all(axis=1))[0], axis=0
                    )

            if len(spots) > 0:
                nb_spot_detected += 1
            frame_counted += 1

        # Light spot is considered as detected if at least in MIN_PERCENTAGE_LIGHTSPOT % of frames
        if frame_counted > 0:
            percentage_spot_detected = nb_spot_detected / frame_counted
            spot_detected = (
                percentage_spot_detected >= min_percentage_light_spot
            )
        else:
            spot_detected = False

        if spot_detected and print_enabled:
            print("nb_spot_detected: ", nb_spot_detected)
            print("frame_counted: ", frame_counted)
            print("spot_detected: ", spot_detected)
            print(
                f"Track: {self.id}_{self.mother_track_id}_to_{','.join(str(daughter) for daughter in self.daughter_track_ids)}"
            )

        return spot_detected

    def get_bridge_images(
        self, video: np.ndarray, margin: int
    ) -> list[np.array]:
        """Generate list of crops around the mid-body.
        First frame is the maximum of cytokinesis frame and mid-body first frame.
        Last frame is the last mid-body frame.

        Parameters
        ----------
        video: np.ndarray
            TYXC
        margin: int
            number of pixels to keep around the mid-body, in all directions

        Returns
        ----------
        bridge_images: list[np.array]
            list of crops around the mid-body, TCYX
        frames: list[int]
            list of frames corresponding to each crop
        """

        # Case with no mid_body detected
        if not self.mid_body_spots:
            return [], []

        ordered_mb_frames = sorted(self.mid_body_spots.keys())
        first_mb_frame = ordered_mb_frames[0]
        last_mb_frame = ordered_mb_frames[-1]

        bridge_images, frames = [], []
        for frame in range(first_mb_frame, last_mb_frame + 1):
            min_x = self.position.min_x
            min_y = self.position.min_y

            # Get midbody coordinates
            frame_mid_body = self.mid_body_spots[frame]
            x_pos, y_pos = min_x + frame_mid_body.x, min_y + frame_mid_body.y

            # Extract frame image and crop around the midbody Sir-tubulin
            frame_image = (
                video[frame, :, :, :].squeeze().transpose(2, 0, 1)
            )  # CYX
            crop = smart_cropping(
                frame_image, margin, x_pos, y_pos, pad=True
            )  # CYX
            bridge_images.append(crop)
            frames.append(frame)

        return bridge_images, frames

    def get_mid_body_legend(
        self,
    ) -> dict[int, dict[str, Union[int, str]]]:
        """Get legend for mid-body spot.

        Returns
        -------
        mid_body_legend : dict[int, dict[str, Union[int, str]]]
            Dictionary with frame number as key and dictionary with x, y and category as value.
        """

        # Check that cut detection was possible
        cut_detection = self.key_events_frame["first_mt_cut"]

        mid_body_legend = {}
        for frame, mid_body_spot in self.mid_body_spots.items():
            # Get category
            if cut_detection < 0:
                mid_body_category = ImpossibleDetection(cut_detection).name
            else:
                mid_body_category = "undefined"
                for category, category_frame in self.key_events_frame.items():
                    if (
                        category_frame > frame
                    ):  # make sure current step is before frame
                        continue
                    if (
                        mid_body_category
                        in self.key_events_frame  # "undefined" case
                        and self.key_events_frame[mid_body_category]
                        > category_frame
                    ):  # make sure we are not going back
                        continue
                    mid_body_category = category
            mid_body_legend[frame] = {
                "x": mid_body_spot.x + self.position.min_x,
                "y": mid_body_spot.y + self.position.min_y,
                "category": snake_to_normal(mid_body_category),
            }
        return mid_body_legend



"""From segmentation mask to list of coordinates.
Most of the code below is inspired from Trackmate (https://github.com/trackmate-sc/TrackMate)"""

import numpy as np


class MaskPolygon:
    """Polygon class to store the coordinates of a polygon."""

    def __init__(self, x, y, count):
        self.x = x
        self.y = y
        self.count = count

    def get_length(self, is_line):
        """Calculate the length of the polygon."""
        length = 0.0
        for i in range(self.count - 1):
            dx = self.x[i + 1] - self.x[i]
            dy = self.y[i + 1] - self.y[i]
            length += np.sqrt(dx * dx + dy * dy)
        if not is_line:
            dx = self.x[0] - self.x[self.count - 1]
            dy = self.y[0] - self.y[self.count - 1]
            length += np.sqrt(dx * dx + dy * dy)
        return length


class Outline:
    """This class implements a Cartesian polygon in progress. The edges are
    supposed to be parallel to the x or y axis. It is implemented as a deque
    to be able to add points to both sides."""

    def __init__(self):
        self.grow = 10
        self.reserved = self.grow
        self.first = int(self.grow / 2)
        self.last = int(self.grow / 2)

        self.x = [0] * self.reserved
        self.y = [0] * self.reserved

    def needs(self, needed_at_begin, needed_at_end):
        """Makes sure that enough free space is available at the beginning and
        end of the list, by enlarging the arrays if required"""
        if (
            needed_at_begin > self.first
            or needed_at_end > self.reserved - self.last
        ):
            extra_space = max(
                self.grow, abs(self.x[self.last - 1] - self.x[self.first])
            )
            new_size = (
                self.reserved + needed_at_begin + needed_at_end + extra_space
            )
            new_first = needed_at_begin + extra_space // 2
            new_x = [0] * new_size
            new_y = [0] * new_size
            new_x[new_first : new_first + (self.last - self.first)] = self.x[
                self.first : self.last
            ]
            new_y[new_first : new_first + (self.last - self.first)] = self.y[
                self.first : self.last
            ]
            self.x = new_x
            self.y = new_y
            self.last += new_first - self.first
            self.first = new_first
            self.reserved = new_size

    def append_outline(self, o):
        """Merge with another Outline by adding it at the end. Thereafter, the
        other outline must not be used any more."""
        size = self.last - self.first
        o_size = o.last - o.first
        if size <= o.first and o_size > self.reserved - self.last:
            # We don't have enough space in our own array but in that of 'o'
            o.x[o.first - size : o.first] = self.x[self.first : self.last]
            o.y[o.first - size : o.first] = self.y[self.first : self.last]
            self.x = o.x
            self.y = o.y
            self.first = o.first - size
            self.last = o.last
            self.reserved = o.reserved
        else:
            # Append to our own array
            self.needs(0, o_size)
            self.x[self.last : self.last + o_size] = o.x[o.first : o.last]
            self.y[self.last : self.last + o_size] = o.y[o.first : o.last]
            self.last += o_size

    def append_point(self, x, y):
        """Adds point x, y at the end of the list"""
        if self.last - self.first >= 2 and collinear(
            self.x[self.last - 2],
            self.y[self.last - 2],
            self.x[self.last - 1],
            self.y[self.last - 1],
            x,
            y,
        ):
            # Replace previous point
            self.x[self.last - 1] = x
            self.y[self.last - 1] = y
        else:
            self.needs(0, 1)  # Ensure space for new point
            self.x[self.last] = x
            self.y[self.last] = y
            self.last += 1

    def prepend_point(self, x, y):
        """Adds point x, y at the beginning of the list"""
        if self.last - self.first >= 2 and collinear(
            self.x[self.first + 1],
            self.y[self.first + 1],
            self.x[self.first],
            self.y[self.first],
            x,
            y,
        ):
            # Replace previous point
            self.x[self.first] = x
            self.y[self.first] = y
        else:
            self.needs(1, 0)  # Ensure space for new point at the beginning
            self.first -= 1
            self.x[self.first] = x
            self.y[self.first] = y

    def prepend_outline(self, o):
        """Merge with another Outline by adding it at the beginning. Thereafter,
        the other outline must not be used any more."""
        size = self.last - self.first
        o_size = o.last - o.first
        if size <= o.reserved - o.last and o_size > self.first:
            # We don't have enough space in our own array but in that of 'o'
            # Append our own data to that of 'o'
            o.x[o.last : o.last + size] = self.x[self.first : self.last]
            o.y[o.last : o.last + size] = self.y[self.first : self.last]
            self.x = o.x
            self.y = o.y
            self.first = o.first
            self.last = o.last + size
            self.reserved = o.reserved
        else:
            # Prepend to our own array
            self.needs(o_size, 0)
            self.first -= o_size
            self.x[self.first : self.first + o_size] = o.x[o.first : o.last]
            self.y[self.first : self.first + o_size] = o.y[o.first : o.last]

    def get_polygon(self):
        """Optimize out intermediate points of straight lines (created, e.g., by merging outlines)"""
        i = self.first + 1
        j = self.first + 1

        while i + 1 < self.last:
            if collinear(
                self.x[j - 1],
                self.y[j - 1],
                self.x[j],
                self.y[j],
                self.x[j + 1],
                self.y[j + 1],
            ):
                # Merge i + 1 into i
                self.last -= 1
            else:
                if i != j:
                    self.x[i] = self.x[j]
                    self.y[i] = self.y[j]
                i += 1
            j += 1

        # Wraparound
        if collinear(
            self.x[j - 1],
            self.y[j - 1],
            self.x[j],
            self.y[j],
            self.x[self.first],
            self.y[self.first],
        ):
            self.last -= 1
        else:
            self.x[i] = self.x[j]
            self.y[i] = self.y[j]

        if self.last - self.first > 2 and collinear(
            self.x[self.last - 1],
            self.y[self.last - 1],
            self.x[self.first],
            self.y[self.first],
            self.x[self.first + 1],
            self.y[self.first + 1],
        ):
            self.first += 1

        count = self.last - self.first
        x_new = self.x[self.first : self.first + count]
        y_new = self.y[self.first : self.first + count]

        return MaskPolygon(x_new, y_new, count)


def collinear(x1, y1, x2, y2, x3, y3) -> bool:
    """Implement the logic to check if three points are collinear

    Parameters
    ----------
    x1 : int
        x-coordinate of the first point.
    y1 : int
        y-coordinate of the first point.
    x2 : int
        x-coordinate of the second point.
    y2 : int
        y-coordinate of the second point.
    x3 : int
        x-coordinate of the third point.
    y3 : int
        y-coordinate of the third point.

    Returns
    -------
    bool
        True if the three points are collinear, False otherwise."""
    return (x2 - x1) * (y3 - y2) == (y2 - y1) * (x3 - x2)


def from_labeling_with_roi(segmentation_mask: np.ndarray) -> list[MaskPolygon]:
    """Convert a segmentation mask to a list of polygons. Each ROI is converted to a polygon.

    Parameters
    ----------
    segmentation_mask : np.ndarray
        Segmentation mask. YX.

    Returns
    -------
    list[MaskPolygon]
        List of polygons. Each polygon is a MaskPolygon object.
    """
    nb_roi = np.max(segmentation_mask)
    polygons = []
    for roi in range(1, nb_roi + 1):
        # Define ROI bounding box
        roi_mask = segmentation_mask == roi
        roi_pixels = np.where(roi_mask)
        roi_min_y = int(np.min(roi_pixels[0]))
        roi_max_y = int(np.max(roi_pixels[0]))
        roi_min_x = int(np.min(roi_pixels[1]))
        roi_max_x = int(np.max(roi_pixels[1]))
        # Extract submask
        sub_mask = roi_mask[roi_min_y:roi_max_y, roi_min_x:roi_max_x]
        # Convert submask to polygon
        local_polygons = mask_to_polygons(sub_mask)
        # Sort them by count and keep the biggest one
        local_polygons.sort(key=lambda x: x.count, reverse=True)
        local_polygon = local_polygons[0]
        # Move polygon to global coordinates
        # NB: Trackmate switches x and y
        local_polygon.x = [lx + roi_min_y for lx in local_polygon.x]
        local_polygon.y = [ly + roi_min_x for ly in local_polygon.y]
        polygons.append(local_polygon)
    return polygons


def mask_to_polygons(mask: np.ndarray):
    """Parse a 2D mask and return a list of polygons for the external contours of white objects.
    Warning: cannot deal with holes, they are simply ignored.
    Copied and adapted from ImageJ1 code by Wayne Rasband."""
    w = mask.shape[0]
    h = mask.shape[1]

    polygons = []
    prev_row = [False] * (w + 2)
    this_row = [False] * (w + 2)
    outline = [None] * (w + 1)

    for y in range(h + 1):

        b = prev_row
        prev_row = this_row
        this_row = b

        x_after_lower_right_corner = -1
        o_after_lower_right_corner = None

        this_row[1] = mask[0, y] if y < h else False

        for x in range(w + 1):

            # we need to read one pixel ahead
            if y < h and x < w - 1:
                this_row[x + 2] = mask[x + 1, y]
            elif x < w - 1:
                this_row[x + 2] = False

            if this_row[x + 1]:  # pixel (x,y) is selected
                if not prev_row[x + 1]:
                    # Upper edge of selected area:
                    # - left and right outlines are null: new outline
                    # - left null: append (line to left)
                    # - right null: prepend (line to right), or
                    # prepend&append (after lower right corner, two borders
                    # from one corner)
                    # - left == right: close (end of hole above) unless we
                    # can continue at the right
                    # - left != right: merge (prepend) unless we can
                    # continue at the right
                    if outline[x] is None:
                        if outline[x + 1] is None:
                            outline[x + 1] = outline[x] = Outline()
                            outline[x].append_point(x + 1, y)
                            outline[x].append_point(x, y)
                        else:
                            outline[x] = outline[x + 1]
                            outline[x + 1] = None
                            outline[x].append_point(x, y)
                    elif outline[x + 1] is None:
                        if x == x_after_lower_right_corner:
                            outline[x + 1] = outline[x]
                            outline[x] = o_after_lower_right_corner
                            outline[x].append_point(x, y)
                            outline[x + 1].prepend_point(x + 1, y)
                        else:
                            outline[x + 1] = outline[x]
                            outline[x] = None
                            outline[x + 1].prepend_point(x + 1, y)
                    elif outline[x + 1] == outline[x]:
                        if (
                            x < w - 1
                            and y < h
                            and x != x_after_lower_right_corner
                            and not this_row[x + 2]
                            and prev_row[x + 2]
                        ):  # at lower right corner & next pxl deselected
                            outline[x] = None
                            outline[x + 1].prepend_point(x + 1, y)
                            x_after_lower_right_corner = x + 1
                            o_after_lower_right_corner = outline[x + 1]
                        else:  # we cannot handle holes
                            outline[x + 1] = None
                            outline[x] = (
                                o_after_lower_right_corner
                                if x == x_after_lower_right_corner
                                else None
                            )
                    else:
                        outline[x].prepend_outline(outline[x + 1])
                        for x1 in range(w + 1):
                            if x1 != x + 1 and outline[x1] == outline[x + 1]:
                                outline[x1] = outline[x]
                                outline[x + 1] = None
                                outline[x] = (
                                    o_after_lower_right_corner
                                    if x == x_after_lower_right_corner
                                    else None
                                )
                                break
                        if outline[x + 1] is not None:
                            raise RuntimeError("Assertion failed")
                if not this_row[x]:
                    if outline[x] is None:
                        raise RuntimeError("Assertion failed")
                    outline[x].append_point(x, y + 1)
            else:  # !thisRow[x + 1], i.e., pixel (x,y) is deselected
                if prev_row[x + 1]:
                    # Lower edge of selected area:
                    # - left and right outlines are null: new outline
                    # - left == null: prepend
                    # - right == null: append, or append&prepend (after
                    # lower right corner, two borders from one corner)
                    # - right == left: close unless we can continue at the
                    # right
                    # - right != left: merge (append) unless we can
                    # continue at the right
                    if outline[x] is None:
                        if outline[x + 1] is None:
                            outline[x] = outline[x + 1] = Outline()
                            outline[x].append_point(x, y)
                            outline[x].append_point(x + 1, y)
                        else:
                            outline[x] = outline[x + 1]
                            outline[x + 1] = None
                            outline[x].prepend_point(x, y)
                    elif outline[x + 1] is None:
                        if x == x_after_lower_right_corner:
                            outline[x + 1] = outline[x]
                            outline[x] = o_after_lower_right_corner
                            outline[x].prepend_point(x, y)
                            outline[x + 1].prepend_point(x + 1, y)
                        else:
                            outline[x + 1] = outline[x]
                            outline[x] = None
                            outline[x + 1].append_point(x + 1, y)
                    elif outline[x + 1] == outline[x]:
                        if (
                            x < w - 1
                            and y < h
                            and x != x_after_lower_right_corner
                            and this_row[x + 2]
                            and not prev_row[x + 2]
                        ):
                            outline[x] = None
                            outline[x + 1].append_point(x + 1, y)
                            x_after_lower_right_corner = x + 1
                            o_after_lower_right_corner = outline[x + 1]
                        else:
                            polygons.append(outline[x].get_polygon())
                            outline[x + 1] = None
                            outline[x] = (
                                o_after_lower_right_corner
                                if x == x_after_lower_right_corner
                                else None
                            )
                    else:
                        if (
                            x < w - 1
                            and y < h
                            and x != x_after_lower_right_corner
                            and this_row[x + 2]
                            and not prev_row[x + 2]
                        ):
                            outline[x].append_point(x + 1, y)
                            outline[x + 1].prepend_point(x + 1, y)
                            x_after_lower_right_corner = x + 1
                            o_after_lower_right_corner = outline[x]
                            outline[x] = None
                        else:
                            outline[x].append_outline(outline[x + 1])  # merge
                            for x1 in range(w + 1):
                                if (
                                    x1 != x + 1
                                    and outline[x1] == outline[x + 1]
                                ):
                                    outline[x1] = outline[x]
                                    outline[x + 1] = None
                                    outline[x] = (
                                        o_after_lower_right_corner
                                        if x == x_after_lower_right_corner
                                        else None
                                    )
                                    break
                            if outline[x + 1] is not None:
                                raise RuntimeError("Assertion failed")
                    if this_row[x]:
                        if outline[x] is None:
                            raise RuntimeError("Assertion failed")
                        outline[x].prepend_point(x, y + 1)

    return polygons


def get_interpolated_polygon(p, interval):
    """Returns an interpolated version of this selection
    with points spaced abs('interval') pixels apart.
    If 'interval' is negative, the program is allowed to decrease abs('interval')
    so that the last segment will hit the end point"""

    allow_to_adjust = interval < 0
    interval = abs(interval)

    length = p.get_length(is_line=False)

    n_points = len(p.x)
    if n_points < 2:
        return p
    if abs(interval) < 0.01:
        raise ValueError("Interval must be >= 0.01")

    p.x.append(p.x[0])
    p.y.append(p.y[0])

    n_points2 = int(10 + (length * 1.5) / interval)  # allow some headroom
    try_interval = interval
    min_diff = 1e9
    best_interval = 0
    src_ptr = 0  # index of source polygon
    dest_ptr = 0  # index of destination polygon
    dest_x_arr = np.zeros(n_points2)
    dest_y_arr = np.zeros(n_points2)
    n_trials = 50
    trial = 0

    while trial <= n_trials:
        dest_x_arr[0] = p.x[0]
        dest_y_arr[0] = p.y[0]
        src_ptr = 0
        dest_ptr = 0
        x_a = p.x[0]  # start of current segment
        y_a = p.y[0]

        while src_ptr < n_points - 1:  # collect vertices
            x_c = dest_x_arr[dest_ptr]  # center circle
            y_c = dest_y_arr[dest_ptr]
            x_b = p.x[src_ptr + 1]  # end of current segment
            y_b = p.y[src_ptr + 1]
            intersections = line_circle_intersection(
                x_a, y_a, x_b, y_b, x_c, y_c, try_interval, True
            )
            if len(intersections) >= 2:
                x_a, y_a = intersections[
                    :2
                ]  # only use first of two intersections
                dest_ptr += 1
                dest_x_arr[dest_ptr] = x_a
                dest_y_arr[dest_ptr] = y_a
            else:
                src_ptr += 1  # no intersection found, pick next segment
                x_a = p.x[src_ptr]
                y_a = p.y[src_ptr]

        dest_ptr += 1
        dest_x_arr[dest_ptr] = p.x[n_points - 1]
        dest_y_arr[dest_ptr] = p.y[n_points - 1]
        dest_ptr += 1
        if not allow_to_adjust:
            break

        n_segments = dest_ptr - 1
        dx = dest_x_arr[dest_ptr - 2] - dest_x_arr[dest_ptr - 1]
        dy = dest_y_arr[dest_ptr - 2] - dest_y_arr[dest_ptr - 1]
        last_seg = np.sqrt(dx * dx + dy * dy)

        diff = last_seg - try_interval
        if abs(diff) < min_diff:
            min_diff = abs(diff)
            best_interval = try_interval

        feedback_factor = (
            0.66  # factor <1: applying soft successive approximation
        )
        try_interval += feedback_factor * diff / n_segments
        # stop if tryInterval < 80% of interval, OR if last segment differs < 0.05 pixels
        if (
            try_interval < 0.8 * interval
            or abs(diff) < 0.05
            or trial == n_trials - 1
        ) and trial < n_trials:
            trial = n_trials
            try_interval = best_interval
        else:
            trial += 1

    # remove closing point from end of array
    dest_ptr -= 1

    x_points_new = np.zeros(dest_ptr)
    y_points_new = np.zeros(dest_ptr)
    for jj in range(dest_ptr):
        x_points_new[jj] = dest_x_arr[jj]
        y_points_new[jj] = dest_y_arr[jj]

    f_poly = MaskPolygon(x_points_new, y_points_new, len(x_points_new))
    return f_poly


def line_circle_intersection(ax, ay, bx, by, cx, cy, rad, ignore_outside):
    """
    Calculates intersections of a line segment with a circle.

    Parameters
    ----------
    ax, ay, bx, by: Points A and B of the line segment.
    cx, cy, rad: Circle center and radius.
    ignore_outside: If true, ignores intersections outside the line segment A-B.

    Returns
    -------
    A list of 0, 2, or 4 coordinates (for 0, 1, or 2 intersection points).
    If two intersection points are returned, they are listed in travel direction A->B.
    """

    # Calculate differences and lengths
    dx_ac = cx - ax
    dy_ac = cy - ay
    len_ac = np.sqrt(dx_ac * dx_ac + dy_ac * dy_ac)

    dx_ab = bx - ax
    dy_ab = by - ay

    # Calculate B2 and C2
    x_b2 = np.sqrt(dx_ab * dx_ab + dy_ab * dy_ab)

    phi1 = np.arctan2(dy_ab, dx_ab)  # amount of rotation
    phi2 = np.arctan2(dy_ac, dx_ac)
    phi3 = phi1 - phi2
    x_c2 = len_ac * np.cos(phi3)
    y_c2 = len_ac * np.sin(phi3)  # rotation & translation is done

    if abs(y_c2) > rad:
        return []  # no intersection found

    half_chord = np.sqrt(rad * rad - y_c2 * y_c2)
    sect_one = x_c2 - half_chord  # first intersection point, still on x axis
    sect_two = x_c2 + half_chord  # second intersection point, still on x axis

    xy_coords = []

    if (0 <= sect_one <= x_b2) or not ignore_outside:
        sect_one_x = (
            np.cos(phi1) * sect_one + ax
        )  # undo rotation and translation
        sect_one_y = np.sin(phi1) * sect_one + ay
        xy_coords.extend([sect_one_x, sect_one_y])

    if (0 <= sect_two <= x_b2) or not ignore_outside:
        sect_two_x = (
            np.cos(phi1) * sect_two + ax
        )  # undo rotation and translation
        sect_two_y = np.sin(phi1) * sect_two + ay
        xy_coords.extend([sect_two_x, sect_two_y])

    if (
        half_chord == 0 and len(xy_coords) > 2
    ):  # tangent line returns only one intersection
        xy_coords = xy_coords[:2]

    return xy_coords


def perpendicular_distance(
    px: float, py: float, vx: float, vy: float, wx: float, wy: float
):
    """Calculate the perpendicular distance from point (px, py) to line segment (vx, vy) - (wx, wy).

    Parameters
    ----------
    px : float
        x-coordinate of the point.
    py : float
        y-coordinate of the point.
    vx : float
        x-coordinate of the start of the line segment.
    vy : float
        y-coordinate of the start of the line segment.
    wx : float
        x-coordinate of the end of the line segment.
    wy : float
        y-coordinate of the end of the line segment.

    Returns
    -------
    float
        Perpendicular distance from the point to the line segment."""
    l2 = (vx - wx) ** 2 + (vy - wy) ** 2  # length squared
    if l2 == 0:
        return np.sqrt((px - vx) ** 2 + (py - vy) ** 2)
    t = ((px - vx) * (wx - vx) + (py - vy) * (wy - vy)) / l2
    t = max(0, min(1, t))
    projection_x = vx + t * (wx - vx)
    projection_y = vy + t * (wy - vy)
    return np.sqrt((px - projection_x) ** 2 + (py - projection_y) ** 2)


def douglas_peucker_core(
    points_list: list[list[float]],
    s: int,
    e: int,
    epsilon: float,
    result_list: list[list[float]],
) -> None:
    """Recursive Douglas-Peucker line simplification.

    Parameters
    ----------
    points_list : list[list[float]]
        List of points to simplify.
    s : int
        Start index.
    e : int
        End index.
    epsilon : float
        Epsilon value for the Douglas-Peucker algorithm.
    result_list : list[list[float]]
        List of simplified points.
    """
    # Find the point with the maximum distance
    dmax = 0
    index = 0

    start = s
    end = e - 1
    for i in range(start + 1, end):
        # Point
        px, py = points_list[i]
        # Start
        vx, vy = points_list[start]
        # End
        wx, wy = points_list[end]
        d = perpendicular_distance(px, py, vx, vy, wx, wy)
        if d > dmax:
            index = i
            dmax = d

    # If max distance is greater than epsilon, recursively simplify
    if dmax > epsilon:
        # Recursive call
        douglas_peucker_core(points_list, s, index, epsilon, result_list)
        douglas_peucker_core(points_list, index, e, epsilon, result_list)
    else:
        if end - start > 0:
            result_list.append(points_list[start])
            result_list.append(points_list[end])
        else:
            result_list.append(points_list[start])


def douglas_peucker(points_list: list[list[float]], epsilon: float):
    """Given a curve composed of line segments find a similar curve with fewer
    points.
    The Ramer–Douglas–Peucker algorithm (RDP) is an algorithm for reducing
    the number of points in a curve that is approximated by a series of
    points.

    Parameters
    ----------
    points_list : list[list[float]]
        List of points to simplify.
    epsilon : float
        Epsilon value for the Douglas-Peucker algorithm.

    Returns
    -------
    list[list[float]]
        Simplified list of points.
    """
    result_list = []
    douglas_peucker_core(
        points_list, 0, len(points_list), epsilon, result_list
    )
    return result_list


def simplify(
    polygon: MaskPolygon, interval: int, epsilon: float
) -> MaskPolygon:
    """Simplify a polygon by reducing the number of points using the Douglas-Peucker algorithm.

    Parameters
    ----------
    polygon : MaskPolygon
        Polygon to simplify.
    interval : int
        Interval between points.
    epsilon : float
        Epsilon value for the Douglas-Peucker algorithm.

    Returns
    -------
    MaskPolygon
        Simplified polygon.
    """
    polygon = get_interpolated_polygon(polygon, interval)
    points = []
    for i in range(polygon.count):
        points.append([polygon.x[i], polygon.y[i]])
    simplified_points = douglas_peucker(points, epsilon)
    x_simplified, y_simplified = zip(*simplified_points)
    # Cast to int
    x_simplified = [int(x) for x in x_simplified]
    y_simplified = [int(y) for y in y_simplified]
    return MaskPolygon(
        list(x_simplified), list(y_simplified), len(x_simplified)
    )


def signed_area(x: list[float], y: list[float]) -> float:
    """Compute the signed area of a polygon.

    Parameters
    ----------
    x : list[float]
        List of x-coordinates of the polygon.
    y : list[float]
        List of y-coordinates of the polygon.

    Returns
    -------
    float
        Signed area of the polygon.
    """
    n = len(x)
    a = 0.0
    for i in range(n - 1):
        a += x[i] * y[i + 1] - x[i + 1] * y[i]
    return (a + x[n - 1] * y[0] - x[0] * y[n - 1]) / 2.0


def centroid(x: list[float], y: list[float]) -> list[float]:
    """Calculate the centroid of a polygon.

    Parameters
    ----------
    x : list[float]
        List of x-coordinates of the polygon.
    y : list[float]
        List of y-coordinates of the polygon.

    Returns
    -------
    list[float]
        Coordinates of the centroid of the polygon.
    """
    area = signed_area(x, y)
    ax = 0.0
    ay = 0.0
    n = len(x)
    for i in range(n - 1):
        w = x[i] * y[i + 1] - x[i + 1] * y[i]
        ax += (x[i] + x[i + 1]) * w
        ay += (y[i] + y[i + 1]) * w

    w0 = x[n - 1] * y[0] - x[0] * y[n - 1]
    ax += (x[n - 1] + x[0]) * w0
    ay += (y[n - 1] + y[0]) * w0

    return [int(ax / 6.0 / area), int(ay / 6.0 / area)]



import os
from skimage import io

from cut_detector._widget import micro_tubules_cut_detection
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.mt_cut_detection import (
    perform_mt_cut_detection,
)


def test_open_micro_tubules_cut_detection_widget():
    """Test opening the micro tubules cut detection widget."""
    micro_tubules_cut_detection()


def test_micro_tubules_cut_detection():
    """Test micro tubules cut detection."""
    video = io.imread(
        os.path.join(get_data_path("videos"), "example_video.tif")  # TYXC
    )

    mitosis_tracks = perform_mt_cut_detection(
        video,
        "example_video",
        get_data_path("mitoses"),
        save=False,
    )

    assert (
        27 <= mitosis_tracks[0].key_events_frame["first_mt_cut"] <= 31
    )  # should be 29
    assert (
        33 <= mitosis_tracks[0].key_events_frame["second_mt_cut"] <= 37
    )  # should be 35



import os
from typing import Optional
import pickle
from skimage import io
import napari

from cut_detector.data.tools import get_data_path
from cut_detector.factories.results_saving_factory import ResultsSavingFactory


def main(
    image_path: Optional[str] = os.path.join(
        get_data_path("videos"), "example_video.tif"
    ),
    mitoses_path: Optional[str] = get_data_path("mitoses"),
):
    # Create a Napari viewer
    viewer = napari.Viewer()

    # Add video
    video = io.imread(image_path)  # TYXC
    # Match Napari video display
    viewer.add_image(video, name="example_video", rgb=True)

    # Load mitosis tracks
    mitosis_tracks = []
    for state_path in os.listdir(mitoses_path):
        with open(os.path.join(mitoses_path, state_path), "rb") as f:
            mitosis_track = pickle.load(f)
        mitosis_tracks.append(mitosis_track)

    ResultsSavingFactory().generate_napari_tracking_mask(
        mitosis_tracks, video, viewer
    )

    napari.run()


if __name__ == "__main__":
    main()



import concurrent.futures
from typing import Optional
import numpy as np
from skimage.morphology import extrema, opening
from scipy import ndimage
from shapely.ops import nearest_points
from shapely import Polygon, Point
from tqdm import tqdm

from ..utils.mid_body_detection.detection import (
    DETECTION_FUNCTIONS,
)
from ..utils.cell_track import CellTrack
from ..constants.tracking import (
    CYTOKINESIS_DURATION,
    MID_BODY_CHANNEL,
    SIR_CHANNEL,
)
from ..utils.mid_body_track import MidBodyTrack
from ..utils.image_tools import smart_cropping
from ..utils.mid_body_spot import MidBodySpot
from ..utils.mitosis_track import MitosisTrack
from ..utils.track_generation import generate_tracks_from_spots
from ..utils.cell_spot import CellSpot
from ..utils.mid_body_detection.tracking import TRACKING_FUNCTIONS


class MidBodyDetectionFactory:
    """
    Class to perform mid-body detection, tracking and filtering.

    Parameters
    ----------
    weight_mklp_intensity_factor : float
        Weight of intensity in spot dist calculation (cf TrackMate).
    weight_sir_intensity_factor : float
        Weight of sir intensity in spot distance calculation.
    mid_body_linking_max_distance :int
        Maximum distance between two mid-bodies to link them.
    cytokinesis_duration : int
        Number of frames to look for mid-body between cells.
    minimum_mid_body_track_length : int
        Minimum spots in mid-body track to consider it.
    """

    def __init__(
        self,
        track_linking_max_distance=175,
        cytokinesis_duration=CYTOKINESIS_DURATION,
        minimum_mid_body_track_length=5,
    ):
        self.track_linking_max_distance = track_linking_max_distance
        self.cytokinesis_duration = cytokinesis_duration
        self.minimum_mid_body_track_length = minimum_mid_body_track_length

    def update_mid_body_spots(
        self,
        mitosis_track: MitosisTrack,
        mitosis_movie: np.ndarray,
        tracks: list[CellTrack],
        parallel_detection: bool,
        detection_method: str,
        tracking_method: str = "spatial_laptrack",
        log_blob_spot: bool = False,
    ) -> None:
        """
        Get spots of best mitosis track.

        Parameters
        ----------
        mitosis_track: MitosisTrack
            Mitosis track to update.
        mitosis_movie: np.ndarray
            Mitosis movie. TYXC.
        tracks: list[CellTrack]
            List of cell tracks.
        parallel_detection: bool
            If True, use parallelization for mid-body spots detection.
        detection_method: str
            Method to detect mid-body spots.
        tracking_method: str
            Method to track mid-body spots.
        log_blob_spot: bool
            If True, display log of spots detected.
        """

        spots_candidates = self.detect_mid_body_spots(
            mitosis_movie,
            method=detection_method,
            parallelization=parallel_detection,
            log_blob_spot=log_blob_spot,
            mitosis_track=mitosis_track,
        )

        assert tracking_method in TRACKING_FUNCTIONS
        mid_body_tracks: list[MidBodyTrack] = generate_tracks_from_spots(
            spots_candidates, TRACKING_FUNCTIONS[tracking_method]
        )

        kept_track = self._select_best_track(
            mitosis_track,
            mid_body_tracks,
            tracks,
            mitosis_movie[..., SIR_CHANNEL],
        )

        if kept_track is None:
            # If necessary, remove mid_body_spots from mitosis_track
            mitosis_track.mid_body_spots = {}
            return

        # Interpolate mid-body spots if gaps in track
        kept_track.fill_gaps()

        # Save spots of best mitosis track
        for rel_frame, spot in kept_track.spots.items():
            frame = rel_frame + mitosis_track.min_frame
            mitosis_track.mid_body_spots[frame] = spot

    def detect_mid_body_spots(
        self,
        mitosis_movie: np.ndarray,
        method: str,
        parallelization: bool = False,
        log_blob_spot: bool = False,
        mitosis_track: Optional[MitosisTrack] = None,
    ) -> dict[int, list[MidBodySpot]]:
        """
        Detect mid-body spots in mitosis movie.

        Parameters
        ----------
        mitosis_movie: np.ndarray
            Mitosis movie. TYXC.
        method: str
            Method to detect mid-body spots.
        parallelization: bool
            If True, use parallelization for mid-body spots detection.
        log_blob_spot: bool
            If True, display log of spots detected.
        mitosis_track: MitosisTrack
            Mitosis track to get mask positions.

        Returns
        ----------
        dict[int, list[MidBodySpot]]
            Dictionary of mid-body spots per frame.
        """
        assert isinstance(parallelization, bool)

        if parallelization:
            return self.thread_pool_detect_mid_body_spots(
                mitosis_movie,
                method,
                mitosis_track,
            )

        return self.serial_detect_mid_body_spots(
            mitosis_movie,
            method,
            log_blob_spot,
            mitosis_track,
        )

    def serial_detect_mid_body_spots(
        self,
        mitosis_movie: np.ndarray,
        method: str,
        log_blob_spot: bool = False,
        mitosis_track: Optional[MitosisTrack] = None,
    ) -> dict[int, list[MidBodySpot]]:
        """Detect mid-body spots in mitosis movie.

        Parameters
        ----------
        mitosis_movie: np.ndarray
            Mitosis movie. TYXC.
        mode: str
            Method to detect mid-body spots.
        log_blob_spot: bool
            If True, display log of spots detected.
        mitosis_track: MitosisTrack
            Mitosis track to get mask positions.
        """

        spots_dictionary = {}
        nb_frames = mitosis_movie.shape[0]

        print("Detecting mid-body spots...")
        for frame in tqdm(range(nb_frames)):

            mitosis_frame = mitosis_movie[frame]  # YXC
            spots = self._spot_detection(
                mitosis_frame,
                method,
                frame,
                log_blob_spot,
                mitosis_track,
            )

            # Update dictionary
            spots_dictionary[frame] = spots

        return spots_dictionary

    def thread_pool_detect_mid_body_spots(
        self,
        mitosis_movie: np.ndarray,
        method: str,
        mitosis_track: Optional[MitosisTrack] = None,
    ) -> dict[int, list[MidBodySpot]]:
        """Detect mid-body spots in mitosis movie.
        Parallelized version.

        Parameters
        ----------
        mitosis_movie: np.ndarray
            Mitosis movie. TYXC.
        mode: str
            Method to detect mid-body spots.
        log_blob_spot: bool
            If True, display log of spots detected.
        mitosis_track: MitosisTrack
            Mitosis track to get mask positions.
        """

        def frame_spot_detection(frame: int) -> tuple[int, list[MidBodySpot]]:
            return frame, self._spot_detection(
                mitosis_movie[frame],
                method,
                frame,
                mitosis_track=mitosis_track,
            )

        future_list = []
        with concurrent.futures.ThreadPoolExecutor() as e:
            for frame in range(mitosis_movie.shape[0]):
                future_list.append(
                    e.submit(
                        frame_spot_detection,
                        frame,
                    )
                )

        return {
            res.result()[0]: res.result()[1]
            for res in concurrent.futures.as_completed(future_list)
        }

    def _spot_detection(
        self,
        image: np.ndarray,
        method: str,
        frame: int,
        log_blob_spot: bool = False,
        mitosis_track: Optional[MitosisTrack] = None,
    ) -> list[MidBodySpot]:
        """Perform spot detection on a single frame of a movie.

        Parameters
        ----------
        image: np.ndarray
            Image to process. YXC.
        method: str
            Method to detect mid-body spots.
        frame: int
            Frame number.
        log_blob_spot: bool
            If True, display log of spots detected.
        mitosis_track: Optional[MitosisTrack]
            Mitosis track to get mask positions.

        Returns
        ----------
        list[MidBodySpot]
            List of detected mid-body spots.
        """

        # Try to load mitosis_track to get mask positions
        if mitosis_track is None:
            image_sir = image[:, :, SIR_CHANNEL]
            image_mklp = image[:, :, MID_BODY_CHANNEL]
            shift_x, shift_y = 0, 0
        elif isinstance(mitosis_track, MitosisTrack):
            mitosis_position = (
                mitosis_track.position
            )  # mitosis position in movie
            frame_position = mitosis_track.dln_positions[
                frame + mitosis_track.min_frame
            ]  # frame position in mitosis
            shift_x = frame_position.min_x - mitosis_position.min_x
            shift_y = frame_position.min_y - mitosis_position.min_y
            image_sir = image[
                shift_y : frame_position.max_y - mitosis_position.min_y,
                shift_x : frame_position.max_x - mitosis_position.min_x,
                SIR_CHANNEL,
            ]
            image_mklp = image[
                shift_y : frame_position.max_y - mitosis_position.min_y,
                shift_x : frame_position.max_x - mitosis_position.min_x,
                MID_BODY_CHANNEL,
            ]
        else:
            raise RuntimeError(
                f"Invalid type for argument mitosis_track: {mitosis_track}"
            )

        if method in DETECTION_FUNCTIONS:
            # Function called referenced by name
            spots = [
                (int(spot[0]), int(spot[1]), int(spot[2]))
                for spot in DETECTION_FUNCTIONS[method](image_mklp)
            ]

            if log_blob_spot:
                for s in spots:
                    print(f"found x:{s[1]}  y:{s[0]}  s:{s[2]}")

        elif method == "h_maxima":  # NB: old method, to be removed
            h_maxima_threshold = 5.0
            # Perform opening followed by closing to remove small spots
            filtered_image = opening(image_mklp, footprint=np.ones((3, 3)))
            # Get local maxima using h_maxima
            local_maxima = extrema.h_maxima(filtered_image, h_maxima_threshold)
            # Label spot regions
            labeled_local_maxima, nb_labels = ndimage.label(
                local_maxima, structure=np.ones((3, 3))
            )
            # Remove inconsistent labels
            # Threshold is computed as 99th percentile of image
            filtering_threshold = np.quantile(image_mklp.flatten(), 0.99)
            for label in range(1, nb_labels + 1):
                # Labels intensity in original image has to be higher than threshold
                if (
                    image_mklp[np.where(labeled_local_maxima == label)].mean()
                    < filtering_threshold
                ):
                    labeled_local_maxima[labeled_local_maxima == label] = 0
            # Re-label accordingly
            labeled_local_maxima, nb_labels = ndimage.label(
                labeled_local_maxima > 0, structure=np.ones((3, 3))
            )
            # Get center of mass to locate spots
            spots = ndimage.center_of_mass(
                local_maxima, labeled_local_maxima, range(1, nb_labels + 1)
            )
            spots = np.asarray(spots, dtype=np.int64)

            # Here, do something to retrieve mid_body area and/or circularity...
            if len(spots) == 0:
                spots = np.array([], dtype=np.int64).reshape(
                    (0, filtered_image.ndim)
                )

        else:
            raise ValueError(f"Unknown mode: [{method}]")

        # Spots can be a list of tuples with 2 or 3 values: 2 values: (y, x) if
        # h_maxima, 3 values: (y, x, sigma) if any blob-based method used
        mid_body_spots = [
            MidBodySpot(
                frame,
                x=position[1] + shift_x,  # switch (y, x) to (x, y)
                y=position[0] + shift_y,
                intensity=self._get_average_intensity(position, image_mklp),
                sir_intensity=self._get_average_intensity(position, image_sir),
            )
            for position in spots
        ]

        return mid_body_spots

    @staticmethod
    def _get_average_intensity(
        position: tuple[int], image: np.ndarray, margin=1
    ) -> int:
        """Get average intensity of a spot in an image.

        Parameters
        ----------
        position : tuple[int]
            Spot position. (y, x).
        image : np.ndarray
            Image to process. YX.
        margin : int
            Margin around the spot to consider.

        Returns
        ----------
        int
            Average intensity of the spot.
        """
        # Get associated crop
        crop = smart_cropping(
            image,
            margin,
            position[1],
            position[0],
            position[1] + 1,
            position[0] + 1,
        )

        # Return average intensity
        return int(np.mean(crop))

    def _get_mid_body_expected_positions(
        self,
        mitosis_track: MitosisTrack,
        cell_tracks: list[CellTrack],
    ) -> tuple[dict, dict[int, CellSpot], dict[int, CellSpot]]:
        """Compute Mid-body expected positions for first cytokinesis frames.
        Defined at the point where the two cells are the closest.
        Outputs are relative to the mitosis_track position.

        Parameters
        ----------
        mitosis_track: MitosisTrack
            Mitosis track to get mother and daughter tracks.
        cell_tracks: list[CellTrack]
            List of cell tracks.

        Returns
        ----------
        tuple[dict, dict[int, CellSpot], dict[int, CellSpot]]
            Expected positions, mother spots, daughter spots.
        """
        (
            mother_track,
            daughter_tracks,
        ) = mitosis_track.get_mother_daughters_tracks(cell_tracks)
        # NB: only first daughter is considered
        daughter_track = daughter_tracks[0]

        rel_expected_positions = {}
        for frame in range(
            daughter_track.start,
            daughter_track.start + self.cytokinesis_duration,
        ):
            # If one cell does not exist anymore, stop
            if (
                frame not in daughter_track.spots
                or frame not in mother_track.spots
            ):
                continue

            # Get relative positions
            rel_positions_mother = [
                [
                    int(mother_point[0]) - mitosis_track.position.min_x,
                    int(mother_point[1]) - mitosis_track.position.min_y,
                ]
                for mother_point in mother_track.spots[frame].spot_points
            ]
            rel_positions_daughter = [
                [
                    int(daughter_point[0]) - mitosis_track.position.min_x,
                    int(daughter_point[1]) - mitosis_track.position.min_y,
                ]
                for daughter_point in daughter_track.spots[frame].spot_points
            ]

            # Try to get points in both cells
            mid_body_candidates = [
                position
                for position in rel_positions_mother
                if position in rel_positions_daughter
            ]
            if (
                len(mid_body_candidates) > 0
            ):  # get mean of mid_body_candidates (as before)
                mid_body_position = np.mean(mid_body_candidates, axis=0)
            else:  # else, get mean of closest_points
                mother_polygon = Polygon(rel_positions_mother)
                daughter_polygon = Polygon(rel_positions_daughter)
                nearest = nearest_points(mother_polygon, daughter_polygon)
                mid_body_point = Point(
                    (nearest[0].x + nearest[1].x) / 2,
                    (nearest[0].y + nearest[1].y) / 2,
                )
                mid_body_position = np.array(
                    [mid_body_point.x, mid_body_point.y]
                )

            rel_expected_positions[frame - mitosis_track.min_frame] = (
                mid_body_position
            )

        return (
            rel_expected_positions,
            mother_track.spots,
            daughter_track.spots,
        )

    def _select_best_track(
        self,
        mitosis_track: MitosisTrack,
        mid_body_tracks: list[MidBodyTrack],
        cell_tracks: list[CellTrack],
        tubulin_movie: np.ndarray,
    ) -> MidBodyTrack:
        """Select best track from mid-body tracks.

        Parameters
        ----------
        mitosis_track: MitosisTrack
            Mitosis track to get mother and daughter tracks.
        mid_body_tracks: list[MidBodyTrack]
            List of mid-body tracks.
        cell_tracks: list[CellTrack]
            List of cell tracks.
        tubulin_movie: np.ndarray
            Tubulin movie. TYX.

        Returns
        -------
        MidBodyTrack
            Best mid-body track.
        """

        # Filter: keep only tracks with sir-tubulin signal at cytokinesis
        # Define range of frames to study
        abs_min_frame = mitosis_track.key_events_frame[
            "cytokinesis"
        ]  # Cytokinesis start
        abs_max_frame = abs_min_frame + int(self.cytokinesis_duration / 2)
        # Randomly sample 1000 intensities to compute threshold
        intensities = np.random.choice(
            tubulin_movie[
                abs_min_frame
                - mitosis_track.min_frame : abs_max_frame
                - mitosis_track.min_frame
                + 1,
                ...,
            ].flatten(),
            1000,
        )
        threshold = np.percentile(intensities, 75)
        # Iterate over all tracks and keep only those with high sir-tubulin signal
        kept_tracks: list[MidBodyTrack] = []
        for track in mid_body_tracks:
            abs_track_frames = [
                frame + mitosis_track.min_frame
                for frame in list(track.spots.keys())
            ]
            # Ignore if no frame in common
            if (
                abs_min_frame > abs_track_frames[-1]
                or abs_max_frame < abs_track_frames[0]
            ):
                continue
            frame_count, total_sir_intensity = 0, 0
            for abs_frame in range(abs_min_frame, abs_max_frame):
                if abs_frame not in abs_track_frames:
                    continue
                frame_count += 1
                track_spot = track.spots[abs_frame - mitosis_track.min_frame]
                total_sir_intensity += tubulin_movie[
                    abs_frame - mitosis_track.min_frame,
                    track_spot.y,
                    track_spot.x,
                ]
            # Ignore if mid-body is not detected in enough frames
            if frame_count < self.minimum_mid_body_track_length:
                continue
            # Ignore if sir-tubulin signal is not high enough
            if total_sir_intensity / frame_count < threshold:
                continue
            kept_tracks.append(track)

        # If no track has sir-tubulin signal, return None
        if len(kept_tracks) == 0:
            return None

        # Sort: choose closest to mid-body expected position
        expected_positions, _, _ = self._get_mid_body_expected_positions(
            mitosis_track, cell_tracks
        )
        # Get list of expected distances
        expected_distances = []
        for track in kept_tracks:
            expected_distances.append(
                track.get_expected_distance(
                    expected_positions,
                    self.track_linking_max_distance,
                )
            )
        assert len(expected_distances) == len(kept_tracks)

        # Sort tracks by expected distance
        sorted_tracks = sorted(
            kept_tracks,
            key=lambda track: expected_distances[kept_tracks.index(track)],
        )
        return sorted_tracks[0]



"""Define optimal tracking functions for every strategy."""

from functools import partial
from typing import Union
from laptrack import LapTrack
import numpy as np
from scipy.spatial import distance

from .spatial_laptrack import SpatialLapTrack


def spatial_intensity_dist(
    c1: tuple,
    c2: tuple,
    max_distance: Union[int, float],
    mklp_weight_factor: float,
    sir_weight_factor: float,
) -> float:
    """Modified version of sqeuclidian distance
    Square Euclidian distance is applied to spatial coordinates x and y.
    An 'intensity' distance is computed with MLKP and SIR intensities.
    Finally, values are combined by weighted addition.

    Parameters
    ----------
    c1 : tuple
        Tuple containing spatial coordinates and intensities of the first point.
    c2 : tuple
        Tuple containing spatial coordinates and intensities of the second point.
    max_distance : Union[int, float]
        Maximum distance for connection.
    mklp_weight_factor : float
        Weight factor for MLKP intensity.
    sir_weight_factor : float
        Weight factor for SIR intensity.

    Returns
    -------
    float
        Distance between two points.
    """

    # unwrapping
    (x1, y1, mlkp1, sir1), (x2, y2, mlkp2, sir2) = c1, c2

    # In case we have a None None point:
    if np.isnan([x1, y1, x2, y2]).any():
        return max_distance * 2  # connection is invalidated

    # spatial coordinates: euclidean
    spatial_e = distance.euclidean([x1, y1], [x2, y2])

    mklp_penalty = (
        3 * mklp_weight_factor * np.abs(mlkp1 - mlkp2) / (mlkp1 + mlkp2)
    )
    sir_penalty = 3 * sir_weight_factor * np.abs(sir1 - sir2) / (sir1 + sir2)
    penalty = 1 + mklp_penalty + sir_penalty
    return (spatial_e * penalty) ** 2


# Distance functions

custom_spatial_distance = partial(
    spatial_intensity_dist,
    max_distance=175,
    mklp_weight_factor=5.0,
    sir_weight_factor=1.50,
)

custom_distance = partial(
    spatial_intensity_dist,
    max_distance=175,
    mklp_weight_factor=5.0,
    sir_weight_factor=1.50,
)

# Tracking methods

TRACKING_FUNCTIONS = {
    "spatial_laptrack": SpatialLapTrack(
        spatial_coord_slice=slice(0, 2),
        spatial_metric="euclidean",
        track_dist_metric=custom_spatial_distance,
        track_cost_cutoff=175,
        gap_closing_dist_metric=custom_spatial_distance,
        gap_closing_cost_cutoff=175,
        gap_closing_max_frame_count=3,
        splitting_cost_cutoff=False,
        merging_cost_cutoff=False,
        alternative_cost_percentile=100,
    ),
    "laptrack": LapTrack(
        track_dist_metric=custom_distance,
        track_cost_cutoff=175**2,
        gap_closing_dist_metric=custom_distance,
        gap_closing_cost_cutoff=175**2,
        gap_closing_max_frame_count=2,
        splitting_cost_cutoff=False,
        merging_cost_cutoff=False,
        alternative_cost_percentile=90,
    ),
}



"""Playground to run cell segmentation."""

import os
from typing import Optional
import matplotlib.pyplot as plt

from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector.data.tools import get_data_path
from cut_detector.models.tools import get_model_path
from cut_detector.factories.segmentation_tracking_factory import (
    SegmentationTrackingFactory,
)


def main(
    image_path: Optional[str] = get_data_path("videos"),
    model_path: Optional[str] = os.path.join(
        get_model_path("segmentation"), "segmentation_model"
    ),
):
    """
    Parameters
    ----------
    image_path : str
        Path to the image to process.
    model_path : str
        Path to the segmentation model.
    """
    # If image_path or model_path are directories, take their first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])
    if os.path.isdir(model_path):
        model_path = os.path.join(model_path, os.listdir(model_path)[0])

    # Read image and preprocess if needed
    image = TiffReader(image_path).image  # TCZYX
    image = image.squeeze()  # TCYX

    # Initialize factory to get constants
    factory = SegmentationTrackingFactory(model_path)

    results, flows, _ = factory.perform_segmentation(image)  # TCYX

    for frame in range(image.shape[0]):
        # Plot image_to_segment and segmented_image
        _, ax = plt.subplots(2, 2)
        ax[0, 0].set_title("Raw image")
        ax[0, 0].imshow(image[frame, 2].squeeze(), cmap="gray")
        ax[0, 1].set_title("Cellpose")
        ax[0, 1].imshow(results[frame], cmap="viridis")
        ax[1, 0].set_title("Flow")
        ax[1, 0].imshow(flows[0][frame])
        ax[1, 1].set_title("Cell probability")
        ax[1, 1].imshow(flows[2][frame], cmap="gray")
        plt.show()


if __name__ == "__main__":
    main()
